try:
	from ldap3 import Server, Connection, ALL, ALL_ATTRIBUTES, SUBTREE
	ldap_available = True
except ModuleNotFoundError:
	ldap_available = False
	Server = None
	Connection = None
	ALL = None
	ALL_ATTRIBUTES = None
	SUBTREE = None

	

# install python ldap3 library:
#   pip install ldap3

# https://ldap3.readthedocs.io/en/latest/searches.html

class Ldap:
	def __init__(self, host, bindDN, password):
		if not ldap_available:
			raise ModuleNotFoundError("Module named ldap3 is not installed. Try: pip install ldap3")
		self._server = Server(host, get_info=ALL)
		self._conn = Connection(self._server, bindDN, password, auto_bind=True)
		self._pageSize = 500
		self._result = []

	def search (self, baseDn, filter="(objectclass=*)", scope=SUBTREE, attributes=ALL_ATTRIBUTES):

		ret = []
		self._conn.search(baseDn, filter, attributes=attributes, 
 					  	  search_scope=scope,
					      paged_size=self._pageSize)
		for e in self._conn.entries:
			ret.append(e) 

		cookie = self._conn.result['controls']['1.2.840.113556.1.4.319']['value']['cookie']
		while cookie:	
			self._conn.search(baseDn, filter,  attributes=attributes, 
						search_scope=scope,
						paged_size=self._pageSize,
						paged_cookie = cookie)
			cookie = self._conn.result['controls']['1.2.840.113556.1.4.319']['value']['cookie']
			for e in self._conn.entries:
				ret.append(e) 

		return ret 


#ldap = Ldap ('ads.company1.com', "cn=Administrator,cn=users,dc=xxx,dc=yyy,dc=zz", "...")
#users = ldap.search ("cn=users,dc=zzz,dc=yyy,dc=zz", 
#					 filter="(objectclass=person)",
#					 attributes = ['cn', 'distinguishedName', 'memberOf', 'sAMAccountName'])



from .base import * 
from .group import *
from .semmgm import *
from .radius import *
from .traceloggers import *
from .cached_property import cached_property

class NcpSemApi(BaseApi):

	def __init__(self, host, *args, tls=True, port=None, **kwargs):
		__doc__ = BaseApi.__init__.__doc__
		self._rootGrpId = 0 
		self._semGroupsHandler = None
		self._host = NcpSemApi._composeURL(host, port, tls)
		self._authCallback = self._getSessionInfo
		BaseApi.__init__(self, host, *args, tls=tls, **kwargs)
		self._host = self._host + "/api/"	# in BaseApi.__init__ auth needs the host without "/api/". But for all future calls we need "/api/" in the URL.

	def _composeURL(host:str, port:int, tls:bool):
		""" When no port is given per host-string or port option,
			append the sem default ports
		"""
		url = BaseApi.composeURL(host, port, tls)
		if len(url.split(":")) == 2 and port is None:
			if tls:
				url += ":12512"
			else:
				url += ":12511"
		return url

	def _getSessionInfo(self):
		ret = self.makeRequest ("GET", "session-info")
		self._rootGrpId	= ret['Result']['RootGroupId']
		self._semGroupsHandler = SemGroupsHandler(self, self._rootGrpId)

	def getRootGroupId(self):
		"""returns id of the root group"""
		if self._rootGrpId == 0:
			self._getSessionInfo()
		return self._rootGrpId

	def getRootGroup(self):
		"""returns the handler of the root group."""
		if self._rootGrpId == 0:
			self._getSessionInfo()
			if self._rootGrpId == 0:
				raise ApiException("Login required")
		entry = SemGroup(self._semGroupsHandler)
		entry.Id = self._rootGrpId
		return entry
		
	def getGroupByName(self, name):
		"""returns the subgroup with the given name, if it exists."""
		if self._rootGrpId == 0:
			self._getSessionInfo()
			if self._rootGrpId == 0:
				raise ApiException("Login required")
		entry = self._semGroupsHandler.find(name)
		return entry

	def getAllGroups(self, rootGroup=None):
		"""recursive iterate over all groups.
			if rootGroup is given, yield only that group and all it's subgroups.
			if rootGroup is None (default), yield all available groups
		"""
		if rootGroup:
			grp = rootGroup
		else:
			grp = self.getRootGroup()
		yield grp
		for subGroup in grp.subGroups:
			yield from self.getAllGroups(subGroup)

	# SEM MGM
	@cached_property
	def semInfo(self):
		"""returns the semInfo handler."""
		return BaseGetHandler(self, "sem-mgm/info")

	@cached_property
	def semSettings(self):
		"""returns the semSettings handler."""
		return BaseListModifyHandler(self, "sem-mgm/settings")
		
	@cached_property
	def semLicense(self):
		"""returns the semLicense handler."""
		return BaseModifyHandler(self, "sem-mgm/license")

	@cached_property
	def stateDBReplication(self):
		"""returns the stateDBReplication handler."""
		return BaseGetHandler(self, "sem-mgm/state-db-repl")

	@cached_property
	def backupServers(self):
		"""returns the backupServers handler."""
		return BaseListHandler(self, "sem-mgm/backup-servers")

	@cached_property
	def semTraceLogs(self):
		"""returns the semTraceLogs handler."""
		return BaseListHandler(self, "sem-mgm/sem-trace-logs")

	@cached_property
	def semTraceLoggers(self):
		"""returns the semTraceLoggers handler."""
		return SemTraceLoggersHandler(self, "sem-mgm/trace-loggers")

	@cached_property
	def radiusClients(self):
		"""returns the radiusClients handler."""
		return RadiusClientsHandler(self, "radius-mgm/clients")
	
	@cached_property
	def radiusConfigurations(self):
		"""returns the radiusConfigurations handler."""
		return RadiusConfigurationsHandler(self, "radius-mgm/configurations")

	@cached_property
	def radiusDictionaries(self):
		"""returns the radiusDictionaries handler."""
		return RadiusDictionariesHandler(self, "radius-mgm/dictionaries")


	def import_script(self, sem_script, group=None):
		"""Import a python script from the SEM and return the loaded module.
			example:
				my_script = NcpSemApi.import_script("my_script")
			produces a similar outcome as 'import my_script' would, if my_script was a file.

			iIf a NcpSemApi.group is given, the script if seached in that group instead of the root group.
			You may also call grp.import_script(...) or give a NcpSemApi.SemScript directly as first parameter.

			When NcpSemApi.import_script() is called, the script is downloaded from the sem into a temporary file.
			The file is imported as module, and gets deleted when the function returns.
		"""
		if group is not None:
			return group.import_script(sem_script)
		else:
			return self.getRootGroup().import_script(sem_script)





from .base import *
from .cached_property import cached_property

class SemTraceLoggersHandler(BaseListUpdateHandler):
	'''SEM Trace Loggers

	Methods
	-------
		createEntry()
			Creates a new SemTraceLogger entry object.
		update(data)

	Inherited Methods
	-----------------
		list (BaseListHandler)
	'''


	def createEntry(self):
		'''Creates a new SemTraceLogger entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return SemTraceLogger(self)

	def update(self, id_or_data : Union[int, BaseEntry, dict], data : Union[None, BaseEntry, dict] = None, options=None) -> Optional[BaseEntry]:
		"""wrapper around BaseUpdateHandler._update(), returns Entry object or None.
			update can be calles eigther:
				handler.update(data)
			if data contains 'Id'
			or:
				handler.update(id, data)
		"""
		if data != None:
			d = data
			assert id_or_data is None, "Id and data was given, but SemTraceLogger does not support Id"
		else:
			d = id_or_data
			if not isinstance(d, BaseEntry) or isinstance(d, dict):
				raise TypeError("update() is missing one required argument: 'data'")
		return self._update(self._url + "/", {"Loggers": [d]}, options)
		# give list(d) to _update, since SemTraceLoggers only accepts arrays


class SemTraceLogger(ModifiableListEntry):
	'''Parameters of a SEM trace logger entry

	Attributes [writable]
	---------------------
		Process : string
			Process
		Logger : string
			Logger
		LogLevel : integer
			Log Level
			Enum Values: 1=error, 2=warning, 3=info, 4=debug, 5=trace, 6=default
	'''

	def __init__(self, getHandler):
		ModifiableListEntry.__init__(self, getHandler)		#Default values
		self._defaultValues = {
			"Process" : "",
			"Logger" : "",
			"LogLevel" : 0,
		}

	def update(self, data=None):
		if data == None:
			_data = self.__dict__["_changes"]
		else:
			_data = data
		_data["Logger"] = self.__dict__["Logger"]
		_data["Process"] = self.__dict__["Process"]
		return self._getHandler.update(None, _data)

	@cached_property
	def Process(self):
		return self.__getattr__("Process")

	@cached_property
	def Logger(self):
		return self.__getattr__("Logger")

	@cached_property
	def LogLevel(self):
		return self.__getattr__("LogLevel")





import urllib.parse
import urllib.request
import urllib.error
import ssl
import json
import base64
import binascii
from .base import ApiException, ApiAuthenticationException

class AuthBase:
	def __init__(self):
		pass

	# this init function is called from BaseApi
	def init(self, host, tls, ssl_verify, ssl_ca_options, callback=None):
		self._host = host
		self._tls = tls 
		self._ssl_verify = ssl_verify
		self._ssl_ca_options = ssl_ca_options
		self._callback = callback # called on every login

	def __del__(self):
		try:
			self._logout()
		except:
			#ingore all errors; and do not print them
			pass

	# derived classes may implement this method
	def getHeader(self):
		"""Returns header-fields needed for the authentication as key-value pairs in a dict. May call _login()"""
		return {}

	# derived classes may implement this method
	def _login(self):
		pass

	# derived classes may implement this method
	def _logout(self):
		pass


class AuthAccessToken(AuthBase):
	def __init__(self, token, type="Bearer"):
		AuthBase.__init__(self)
		self.setToken(token, type)
		self._oauthTokenUrlPath  = "/oauth2/token" 
		self._oauthRevokeUrlPath = "/oauth2/revoke" 

	def setToken(self, token, type):
		self._accessToken = token 
		self._tokenType = type

	def getHeader(self):
		if self._accessToken and self._tokenType:
			return {'Authorization': self._tokenType + " " + self._accessToken}
		return {}

class AuthClientCredential(AuthAccessToken):
	"""oauth2 client credentioal authentication.
		Needs a clientId and a clientSecret.
	"""

	def __init__(self, clientId, clientSecret):
		AuthAccessToken.__init__(self, "")
		self._clientId = clientId
		self._clientSecret = clientSecret

	def getHeader(self):
		if not self._accessToken:
			self._login() 
		return AuthAccessToken.getHeader(self)

	def _sendRequest(self, url, data, headers, method):
		resp = ""	# json http response as dict
		err = None	# exception
		try:
			if self._tls:
				context = ssl.create_default_context(**self._ssl_ca_options)
				if not self._ssl_verify:
					context.check_hostname = False
					context.verify_mode = ssl.CERT_NONE
			else:
				context = None
			conn = urllib.request.Request(url, data=data, headers=headers, method=method)
			res = urllib.request.urlopen(conn, context=context)
			content = res.read().decode("utf-8")
			if content:
				resp = json.loads(content)
		except urllib.error.HTTPError as e:
			if e.code == 502:
				raise ApiException ("HTTP Error 502: Server is not running")

			content = e.read().decode("utf-8")
			if content:
				resp = json.loads(content)	#when this fails, it might be an ssl error...
				e.msg = resp["error"] + ": " + resp["error_description"]
			err = e

		finally:
			if err:
				if err.code == 400:
					raise ApiAuthenticationException("HTTP Error 400: "+err.msg)
				else:
					raise err
		return resp

	def _login(self, verbose=False, show_requests=False):
		headers = {
			'Accept'		: 'application/json,UTF-8',
			'Content-Type'	: 'application/x-www-form-urlencoded',
		}
		url = self._host + self._oauthTokenUrlPath
		data = f"client_id={urllib.parse.quote(self._clientId)}"
		data += f"&client_secret={urllib.parse.quote(self._clientSecret)}"
		data +=	"&grant_type=client_credentials"
		data = bytes(data, encoding="utf-8")

		if verbose:
			print ("Login ------------")
		if verbose or show_requests:
			print ("POST " + url)
		if verbose:
			print ("\t"+ str(headers))
			print ("\t"+ str(data)) # this prints credentials

		resp = self._sendRequest(url, data, headers, "POST")
		self.setToken(resp.get("access_token"), resp.get("token_type"))

		if verbose:
			print(resp)
		if self._callback:
			self._callback()
		
	def __del__(self):
		try:
			self._logout()
		except:
			#ingore all errors; and do not print them
			pass

	def _logout(self, verbose=False):
		if not self._accessToken:
			return

		authorization_b64 = base64.b64encode("{}:{}".format(self._clientId, self._clientSecret).encode()).decode('ascii')
		headers = {
			'Accept'		: 'application/json,UTF-8',
			'Content-Type'	: 'application/x-www-form-urlencoded',
			'Authorization': "Basic " + authorization_b64 
		}
		url = self._host + self._oauthRevokeUrlPath
		data = f"token={urllib.parse.quote(self._accessToken)}"
		data += "&token_type_hint=access_token"
		data = bytes(data, encoding="utf-8")

		if verbose:
			print ("Logout -----------")
			print ("POST " + url)
			print ("\t"+ str(headers))
			print ("\t"+ str(data)) # this prints credentials
		resp = self._sendRequest(url, data, headers, "POST")
		if verbose:
			print(resp)



from .base import *
from .client import *
from .firewall import *
from .radius import *
from .pki import *
from .license import *
from .semmgm import *

from .cached_property import cached_property

import importlib.util
import sys
import tempfile

class SemGroup(LazyListEntry):
	'''Configuration of a sem group

	Attributes [read-only]
	----------------------
		ModifiedOn : time
			Modified on
		ModifiedBy : string
			Modified by

	Attributes [writable]
	---------------------
		Name : string
			Name

		Description : string
			Description

		FullGroupName : string
			Full group name

		Suffix : string
			Suffix

		LdapDN : string
			LDAP DN

		InitUser : string
			Init Username

		UseLicKeysInSubGroups : boolean
			Use license keys in subgroups

		DebugLevel : integer
			Debug Level

		MaxManagedUnits : integer
			Max. managed units

		AuthCodeResetClientCfg : integer
			Reset Authentication Code: Client Configuration
			Enum Values: 0=default, 1=disableReset, 2=defaultWithSubgroups, 3=disableResetWithSubgroups

		AuthCodeResetCert : integer
			Reset Authentication Code: Certificate
			Enum Values: 0=default, 1=disableReset, 2=defaultWithSubgroups, 3=disableResetWithSubgroups

	Sub-Handlers
	------------
		subGroups
			Access SemGroupsHandler

		softwareUpdateLists
			Access SoftwareUpdateListsHandler

		admins
			Access AdminsHandler

		adminGroups
			Access AdminGroupsHandler

		notifications
			Access NotificationsHandler

		semLogs
			Access SemLogsHandler

		tasks
			Tasks TasksHandler

		clientTemplates
			Access ClientTemplatesHandler
			Configuration of Client Templates

		clientConfigurations
			Access ClientConfigurationsHandler
			Configuration of Clients

		firewallTemplates
			Access FirewallTemplatesHandler
			Configuration of Firewall Templates

		radiusTemplates
			Access RadiusTemplatesHandler
			Configuration of RADIUS Templates

		radiusUserConfigurations
			Access RadiusUserConfigurationsHandler
			Configuration of RADIUS Users

		radiusGroupSettings
			Access RadiusGroupSettingsHandler
			Configuration of RADIUS Groups Settings

		radiusUsers
			Access RadiusUsersHandler
			State of RADIUS Users

		radiusSessions
			Access RadiusSessionsHandler
			State of RADIUS Sessions

		radiusAccountingDetails
			Access RadiusAccountingDetailsHandler
			RADIUS Accounting Details

		licenseDetails
			Access LicensesHandler
			Management of licenses

		licenseOverview
			Access LicenseOverviewHandler
			Overview of licenses

		subscriptions
			Access SubscriptionsHandler
			Management of subscriptions

		certificateRequests
			Access CertificateRequestsHandler
			Mamagement of Certificarte Requests

		issuedCertificates
			Access IssuedCertificatesHandler
			Mamagement of issued Requests

		certificateTemplates
			Access CertificateTemplatesHandler
			Configuration of PKI Certificate Templates

		pkiProfiles
			Access PkiProfilesHandler
			Configuration of PKI Profiles
	'''

	#--------------------------------------------------------
	# SEM Base Management
	#--------------------------------------------------------
	@cached_property
	def subGroups(self):
		return SemGroupsHandler(self._getHandler._api, self.Id)

	@cached_property
	def softwareUpdateLists(self):
		return SoftwareUpdateListsHandler(self._getHandler._api, self.Id)

	@cached_property
	def admins(self):
		return AdminsHandler(self._getHandler._api, self.Id)

	@cached_property
	def adminGroups(self):
		return AdminGroupsHandler(self._getHandler._api,self.Id)

	@cached_property
	def notifications(self):
		return NotificationsHandler(self._getHandler._api, self.Id)

	@cached_property
	def semLogs(self):
		return SemLogsHandler(self._getHandler._api, self.Id)

	@cached_property
	def semScripts(self):
		return SemScriptsHandler(self._getHandler._api, self.Id)

	@cached_property
	def tasks(self):
		return TasksHandler(self._getHandler._api, self.Id)

	def import_script(self, sem_script):
		"""Import a python script from the SEM and return the loaded module.
			examples:
				my_script = grp.import_script("my_script")
			or
				script = grp.semScripts.find("my_script")
				my_script = grp.import_script(script)
			both examples produce a similar outcome as 'import my_script' would, if my_script was a file.

			When import_script() is called, the script is downloaded from the SEM into a temporary file.
			The file is imported as module, and gets deleted when the function returns.
		"""
		if isinstance(sem_script, str):
			sem_script = self.semScripts.find(sem_script, throw=True)	# raises Exception if not found
		if not isinstance(sem_script, SemScript):
			raise TypeError(f"First argument of import_script() must be of type SemScript or str, but is of type {type(sem_script)}")

		delete_args = {
			"delete": True
		}
		if sys.version_info.minor >= 12:
			# since python 3.12 delete_on_close is needed on windows, but it is not defined on earlier python versions
			delete_args["delete_on_close"] = False

		with tempfile.NamedTemporaryFile(mode="w", suffix=".py", encoding="utf-8", **delete_args) as file:
			script = sem_script.Script
			# the encoding of script from sem is latin-1, but it is stored in a string, so python assumes it is utf-8 encoded.
			# this line converts it to the correct utf-8 string:
			script = bytes(script, "latin-1").decode("latin-1")
			file.write(script)
			file.seek(0)
			path = file.name
			# for documentation see: https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
			spec = importlib.util.spec_from_file_location(sem_script.Name, path)
			module = importlib.util.module_from_spec(spec)
			sys.modules[sem_script.Name] = module
			spec.loader.exec_module(module)
			return module


	#--------------------------------------------------------
	# Client plug-in
	#--------------------------------------------------------
	# Configuration of Client Templates
	@cached_property
	def clientTemplates(self):
		return ClientTemplatesHandler (self._getHandler._api, self.Id)

	# Configuration of Clients
	@cached_property
	def clientConfigurations(self):
		return ClientConfigurationsHandler (self._getHandler._api, self.Id)


	#--------------------------------------------------------
	# Firewall plug-in
	#--------------------------------------------------------
	# Configuration of Firewall Templates
	@cached_property
	def firewallTemplates(self):
		return FirewallTemplatesHandler(self._getHandler._api, self.Id)

	#--------------------------------------------------------
	# RADIUS plug-in
	#--------------------------------------------------------
	# Configuration of RADIUS Templates
	@cached_property
	def radiusTemplates(self):
		return RadiusTemplatesHandler (self._getHandler._api, self.Id)

	# Configuration of RADIUS Users
	@cached_property
	def radiusUserConfigurations(self):
		return RadiusUserConfigurationsHandler (self._getHandler._api, self.Id)

	# Configuration of RADIUS Groups Settings
	@cached_property
	def radiusGroupSettings(self):
		return RadiusGroupSettingsHandler(self._getHandler._api, self.Id)

	# State of RADIUS Users
	@cached_property
	def radiusUsers(self):
		return RadiusUsersHandler(self._getHandler._api, self.Id)

	# State of RADIUS Sessions
	@cached_property
	def radiusSessions(self):
		return RadiusSessionsHandler(self._getHandler._api, self.Id)

	# RADIUS Accounting Details
	@cached_property
	def radiusAccountingDetails(self):
		return RadiusAccountingDetailsHandler(self._getHandler._api, self.Id)

	#--------------------------------------------------------
	# License Plug-in
	#--------------------------------------------------------
	# Management of licenses
	@cached_property
	def licenseDetails(self):
		return LicenseDetailsHandler(self._getHandler._api, self.Id)

	# Overview of licenses
	@cached_property
	def licenseOverview(self):
		return LicenseOverviewHandler(self._getHandler._api, self.Id)

	# Management of subscriptions
	@cached_property
	def subscriptions(self):
		return SubscriptionsHandler(self._getHandler._api, self.Id)

	#--------------------------------------------------------
	# PKI plug-in
	#--------------------------------------------------------
	# Mamagement of Certificarte Requests
	@cached_property
	def certificateRequests(self):
		return CertificateRequestsHandler(self._getHandler._api, self.Id)

	# Mamagement of issued Requests
	@cached_property
	def issuedCertificates(self):
		return IssuedCertificatesHandler(self._getHandler._api, self.Id)

	# Mamagement of CA certificates
	@cached_property
	def cACertificates(self):
		return CACertificatesHandler(self._getHandler._api, self.Id)

	# Configuration of PKI Certificate Templates
	@cached_property
	def certificateTemplates(self):
		return CertificateTemplatesHandler(self._getHandler._api, self.Id)

	# Configuration of PKI Profiles
	@cached_property
	def pkiProfiles(self):
		return PkiProfilesHandler(self._getHandler._api, self.Id)


	#--------------------------------------------------------
	# Attributes
	#--------------------------------------------------------

	@cached_property
	def Name(self):
		return self.__getattr__("Name")

	@cached_property
	def Description(self):
		return self.__getattr__("Description")

	@cached_property
	def FullGroupName(self):
		return self.__getattr__("FullGroupName")

	@cached_property
	def Suffix(self):
		return self.__getattr__("Suffix")

	@cached_property
	def LdapDN(self):
		return self.__getattr__("LdapDN")

	@cached_property
	def InitUser(self):
		return self.__getattr__("InitUser")

	@cached_property
	def UseLicKeysInSubGroups(self):
		return self.__getattr__("UseLicKeysInSubGroups")

	@cached_property
	def DebugLevel(self):
		return self.__getattr__("DebugLevel")

	@cached_property
	def MaxManagedUnits(self):
		return self.__getattr__("MaxManagedUnits")

	@cached_property
	def AuthCodeResetClientCfg(self):
		return self.__getattr__("AuthCodeResetClientCfg")

	@cached_property
	def AuthCodeResetCert(self):
		return self.__getattr__("AuthCodeResetCert")


class SemGroupsHandler(BaseListModifyHandler):
	def __init__ (self, api, parentGrpId):
		BaseHandler.__init__ (self, api, "sem-mgm/groups")
		self._parentGrpId = parentGrpId

	def list (self, offset=0, limit=10000, sortKey=None, sortDir=None, searchFilter=None):
		# offset has to be set manually if sem has over 10000 groups
		url = self._url + "/" + str(self._parentGrpId) + "/subgroups"
		return self._list (url, offset, limit, sortKey, sortDir, searchFilter)

	def find (self, groupName, options={}):
		# find uses the absolute FullGroupName - so it is not relavtive to the group it it called on!
		url = self._url + "/" + str(self._parentGrpId) + "/find"
		modifiedOptions = copy.deepcopy(options)
		if "returnValues" not in options:
			modifiedOptions["returnValues"] = "false"
		return self._find (url, FullGroupName=groupName, **modifiedOptions)

	def insert (self, data, options=None):
		url = self._url + "/" + str(self._parentGrpId) + "/subgroups"
		return self._insert(url, data, options)

	def createEntry(self) :
		return SemGroup (self)





#---------------------------------------------------------------
# This file is generated! Dont make any changes in this file.
#---------------------------------------------------------------
from .base import *
from .cached_property import cached_property



class FirewallTemplatesHandler(BaseListFindHandler, BaseListUpdateHandler, BaseListInsertHandler, BaseListDeleteHandler):
	'''Management of firewall templates

	Methods
	-------
		createEntry()
			Creates a new FirewallTemplate entry object.
	
	Inherited Methods
	-----------------
		delete (BaseListDeleteHandler)
		find (BaseListFindHandler)
		get (BaseListGetHandler)
		insert (BasListInsertHandler)
		list (BaseListHandler)
		update (BaseListUpdateHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "fw-mgm/{groupid}/templates".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid

	def createEntry(self):
		'''Creates a new FirewallTemplate entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return FirewallTemplate(self, self._groupid)


class FirewallTemplate(LazyModifiableListEntry):
	'''Firewall Template

	Attributes [read-only]
	----------------------
		ModifiedOn : time
			Modified on
		ModifiedBy : string
			Modified by
		ConfiguredIn : string
			Configured In

	Attributes [writable]
	---------------------
		Name : string
			Name
		TemplateType : integer
			Template Type
			Enum Values: 1=Enterprise, 2=Telekom, 3=NCP Exclusive Remote Access Client, 4=VS GovNet Connector 2.0, 5=VS GovNet Connector >=2.1
		Activate : boolean
			Enable Firewall
		PermitIPsec : boolean
			Permit IPsec protocol
		FndNetworks : array from model FndNetwork
			FND Networks
		FndMethod : integer
			FND Method
			Enum Values: 0=eap, 1=tls
		FndMode : integer
			IP address automatically over DHCP assigned
			Enum Values: 0=manually, 1=automatic, 2=DHCP
		FndServer : string
			IP address of FND Server
		FndUserId : string
			User ID
		FndPassword : string
			Password
		FndCertSubject : string
			Incoming Certificate's Subject
		FndCertFingerprint : string
			Issuers Certificate's Fingerprint
		FndPeriodicallyTimer : integer
			Check for friendly net detection periodically
		VpnDenyInFnd : boolean
			Connection set up not permitted in detected friendly network
		HideCredentialProvFnd : boolean
			Mask out logon options in detected friendly network
		FndVpnTimeout : integer
			Timeout for FND before Windows logon
		FndActions : array from model FndAction
			Actions
		StatefulBootOption : integer
			Enable Stateful Boot Option
			Enum Values: 0=unlocked, 1=locked, 2=Statefull boot option
		UdpPreFiltering : integer
			UDP Pre-filtering
			Enum Values: 0=off, 1=always, 2=fnd off
		VmGuestLock : boolean
			Protect VMware guest operating systems
		OutgoingDropReset : boolean
			Reject outgoing traffic
		HomeZoneMode : boolean
			Enable Home Zone
		HomeZoneExitOnLinkDown : boolean
			Exit Home Zone on link status down for more than 10s.
		LogLevel : integer
			Enable firewall log
			Enum Values: 0=off, 2=log drop, 4=log accept, 6=log accept and drop
		LogHoldTime : integer
			Days of logging
		LogPath : string
			Path for log files
		InheritedToSubgroups : boolean
			Entry inherited to subgroups

	Sub-Handlers
	------------
		firewallRules
			Access FirewallRulesHandler
	'''

	def __init__(self, getHandler, groupid):
		LazyModifiableListEntry.__init__(self, getHandler)
		self._groupid = groupid
		self._api = getHandler._api
		# Default values
		self._defaultValues = {
			"Name" : "",
			"TemplateType" : 0,
			"Activate" : False,
			"PermitIPsec" : True,
			"FndNetworks" : [],
			"FndMethod" : "eap",
			"FndMode" : "manually",
			"FndServer" : "",
			"FndUserId" : "",
			"FndPassword" : "",
			"FndCertSubject" : "",
			"FndCertFingerprint" : "",
			"FndPeriodicallyTimer" : 3600,
			"VpnDenyInFnd" : False,
			"HideCredentialProvFnd" : False,
			"FndVpnTimeout" : 60,
			"FndActions" : [],
			"StatefulBootOption" : "unlocked",
			"UdpPreFiltering" : "off",
			"VmGuestLock" : False,
			"OutgoingDropReset" : False,
			"HomeZoneMode" : False,
			"HomeZoneExitOnLinkDown" : False,
			"LogLevel" : "off",
			"LogHoldTime" : 30,
			"LogPath" : "",
			"ModifiedOn" : "",
			"ModifiedBy" : "",
			"ConfiguredIn" : "",
			"InheritedToSubgroups" : False,
		}

	@cached_property
	def Name(self):
		return self.__getattr__("Name")

	@cached_property
	def TemplateType(self):
		return self.__getattr__("TemplateType")

	@cached_property
	def Activate(self):
		return self.__getattr__("Activate")

	@cached_property
	def PermitIPsec(self):
		return self.__getattr__("PermitIPsec")

	@cached_property
	def FndNetworks(self):
		return self.__getattr__("FndNetworks")

	@cached_property
	def FndMethod(self):
		return self.__getattr__("FndMethod")

	@cached_property
	def FndMode(self):
		return self.__getattr__("FndMode")

	@cached_property
	def FndServer(self):
		return self.__getattr__("FndServer")

	@cached_property
	def FndUserId(self):
		return self.__getattr__("FndUserId")

	@cached_property
	def FndPassword(self):
		return self.__getattr__("FndPassword")

	@cached_property
	def FndCertSubject(self):
		return self.__getattr__("FndCertSubject")

	@cached_property
	def FndCertFingerprint(self):
		return self.__getattr__("FndCertFingerprint")

	@cached_property
	def FndPeriodicallyTimer(self):
		return self.__getattr__("FndPeriodicallyTimer")

	@cached_property
	def VpnDenyInFnd(self):
		return self.__getattr__("VpnDenyInFnd")

	@cached_property
	def HideCredentialProvFnd(self):
		return self.__getattr__("HideCredentialProvFnd")

	@cached_property
	def FndVpnTimeout(self):
		return self.__getattr__("FndVpnTimeout")

	@cached_property
	def FndActions(self):
		return self.__getattr__("FndActions")

	@cached_property
	def StatefulBootOption(self):
		return self.__getattr__("StatefulBootOption")

	@cached_property
	def UdpPreFiltering(self):
		return self.__getattr__("UdpPreFiltering")

	@cached_property
	def VmGuestLock(self):
		return self.__getattr__("VmGuestLock")

	@cached_property
	def OutgoingDropReset(self):
		return self.__getattr__("OutgoingDropReset")

	@cached_property
	def HomeZoneMode(self):
		return self.__getattr__("HomeZoneMode")

	@cached_property
	def HomeZoneExitOnLinkDown(self):
		return self.__getattr__("HomeZoneExitOnLinkDown")

	@cached_property
	def LogLevel(self):
		return self.__getattr__("LogLevel")

	@cached_property
	def LogHoldTime(self):
		return self.__getattr__("LogHoldTime")

	@cached_property
	def LogPath(self):
		return self.__getattr__("LogPath")

	@cached_property
	def ModifiedOn(self):
		return self.__getattr__("ModifiedOn")

	@cached_property
	def ModifiedBy(self):
		return self.__getattr__("ModifiedBy")

	@cached_property
	def ConfiguredIn(self):
		return self.__getattr__("ConfiguredIn")

	@cached_property
	def InheritedToSubgroups(self):
		return self.__getattr__("InheritedToSubgroups")

	@cached_property
	def firewallRules(self):
		'''Returns handler to access FirewallRules'''
		return FirewallRulesHandler(self._api, self._groupid, self.Id)


class FirewallRulesHandler(BaseListFindHandler, BaseListUpdateHandler, BaseListInsertHandler, BaseListDeleteHandler):
	'''Configuration of firewall template rules

	Methods
	-------
		createEntry()
			Creates a new FirewallRule entry object.
	
	Inherited Methods
	-----------------
		delete (BaseListDeleteHandler)
		find (BaseListFindHandler)
		get (BaseListGetHandler)
		insert (BasListInsertHandler)
		list (BaseListHandler)
		update (BaseListUpdateHandler)
	'''
	
	def __init__(self, api, groupid, fwtemplid):
		url = "fw-mgm/{groupid}/templates/{fwtemplid}/rules".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid
		self._fwtemplid = fwtemplid

	def createEntry(self):
		'''Creates a new FirewallRule entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return FirewallRule(self, self._groupid, self._fwtemplid)


class FirewallRule(LazyModifiableListEntry):
	'''Firewall rule

	Attributes [read-only]
	----------------------
		ModifiedOn : time
			Modified on
		ModifiedBy : string
			Modified by

	Attributes [writable]
	---------------------
		Name : string
			Name
		Direction : integer
			Direction
			Enum Values: 0=bidirectional, 1=out, 2=in
		State : boolean
			Action
		LocalIpAddrs : ipaddr.range
			Local IP Addresses
		RemoteIpAddrs : ipaddr.range
			Remote IP Addresses
		LocalPorts : port.range
			Local Ports
		RemotePorts : port.range
			Remote Ports
		TypeVPN : boolean
			Network type VPN
		TypeFriendly : boolean
			Network type friendly
		TypeUnknown : boolean
			Network type unknown
		TypeHomeZone : boolean
			Network type home zone
		Protocol : string
			Protocol
		Application : string
			Application
		NoAutoConnectMobileNetwork : boolean
			No auto connect via mobile network
		ValidNoVpnConnection : boolean
			Only valid with an inactive VPN Connection
	'''

	def __init__(self, getHandler, groupid, fwtemplid):
		LazyModifiableListEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Name" : "",
			"Direction" : "bidirectional",
			"State" : True,
			"LocalIpAddrs" : "",
			"RemoteIpAddrs" : "",
			"LocalPorts" : "",
			"RemotePorts" : "",
			"TypeVPN" : False,
			"TypeFriendly" : False,
			"TypeUnknown" : False,
			"TypeHomeZone" : False,
			"Protocol" : "",
			"Application" : "",
			"NoAutoConnectMobileNetwork" : False,
			"ValidNoVpnConnection" : False,
			"ModifiedOn" : "",
			"ModifiedBy" : "",
		}

	@cached_property
	def Name(self):
		return self.__getattr__("Name")

	@cached_property
	def Direction(self):
		return self.__getattr__("Direction")

	@cached_property
	def State(self):
		return self.__getattr__("State")

	@cached_property
	def LocalIpAddrs(self):
		return self.__getattr__("LocalIpAddrs")

	@cached_property
	def RemoteIpAddrs(self):
		return self.__getattr__("RemoteIpAddrs")

	@cached_property
	def LocalPorts(self):
		return self.__getattr__("LocalPorts")

	@cached_property
	def RemotePorts(self):
		return self.__getattr__("RemotePorts")

	@cached_property
	def TypeVPN(self):
		return self.__getattr__("TypeVPN")

	@cached_property
	def TypeFriendly(self):
		return self.__getattr__("TypeFriendly")

	@cached_property
	def TypeUnknown(self):
		return self.__getattr__("TypeUnknown")

	@cached_property
	def TypeHomeZone(self):
		return self.__getattr__("TypeHomeZone")

	@cached_property
	def Protocol(self):
		return self.__getattr__("Protocol")

	@cached_property
	def Application(self):
		return self.__getattr__("Application")

	@cached_property
	def NoAutoConnectMobileNetwork(self):
		return self.__getattr__("NoAutoConnectMobileNetwork")

	@cached_property
	def ValidNoVpnConnection(self):
		return self.__getattr__("ValidNoVpnConnection")

	@cached_property
	def ModifiedOn(self):
		return self.__getattr__("ModifiedOn")

	@cached_property
	def ModifiedBy(self):
		return self.__getattr__("ModifiedBy")



class FirewallTemplateList(BaseEntry):
	'''Model for firewall template get list operation

	Attributes [writable]
	---------------------
		Id : integer
			REST ID
		Name : string
			Name
		GroupId : integer
			Group ID
		ConfiguredIn : string
			Configured in
		TemplateType : integer
			Template Type
			Enum Values: 1=Enterprise, 2=Telekom, 3=NCP Exclusive Remote Access Client, 4=VS GovNet Connector 2.0, 5=VS GovNet Connector >=2.1
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Name" : "",
			"GroupId" : 0,
			"ConfiguredIn" : "",
			"TemplateType" : 0,
		}

	@cached_property
	def Id(self):
		return self.__getattr__("Id")

	@cached_property
	def Name(self):
		return self.__getattr__("Name")

	@cached_property
	def GroupId(self):
		return self.__getattr__("GroupId")

	@cached_property
	def ConfiguredIn(self):
		return self.__getattr__("ConfiguredIn")

	@cached_property
	def TemplateType(self):
		return self.__getattr__("TemplateType")


class FndNetwork(BaseEntry):
	'''FND Network

	Attributes [writable]
	---------------------
		FndNetwork : ipaddr.range
			IP Network
		FndNetmask : ipaddr.range
			Netmask
		DhcpIpAddrRanges : ipaddr.range
			Dhcp Ip Addr Ranges
		FndDescription : string
			DHCP Server
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"FndNetwork" : "",
			"FndNetmask" : "",
			"DhcpIpAddrRanges" : "",
			"FndDescription" : "",
		}

	@cached_property
	def FndNetwork(self):
		return self.__getattr__("FndNetwork")

	@cached_property
	def FndNetmask(self):
		return self.__getattr__("FndNetmask")

	@cached_property
	def DhcpIpAddrRanges(self):
		return self.__getattr__("DhcpIpAddrRanges")

	@cached_property
	def FndDescription(self):
		return self.__getattr__("FndDescription")


class FndAction(BaseEntry):
	'''FND Action

	Attributes [writable]
	---------------------
		Application : string
			Applications
		Attribute : integer
			Start
			Enum Values: 131072=friendlyNet, 262144=unknown, 8388608=homeZone
		Wait : boolean
			Wait application finished
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Application" : "",
			"Attribute" : "friendlyNet",
			"Wait" : True,
		}

	@cached_property
	def Application(self):
		return self.__getattr__("Application")

	@cached_property
	def Attribute(self):
		return self.__getattr__("Attribute")

	@cached_property
	def Wait(self):
		return self.__getattr__("Wait")


class FirewallRuleList(BaseEntry):
	'''Firewall rule list entry

	Attributes [read-only]
	----------------------
		ConfiguredIn : string
			Configured In

	Attributes [writable]
	---------------------
		FWPOL_C_INDEX : integer
			FWPOL_C_INDEX
		Name : string
			Name
		GroupId : integer
			Group ID
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"FWPOL_C_INDEX" : 0,
			"Name" : "",
			"GroupId" : 0,
			"ConfiguredIn" : "",
		}

	@cached_property
	def FWPOL_C_INDEX(self):
		return self.__getattr__("FWPOL_C_INDEX")

	@cached_property
	def Name(self):
		return self.__getattr__("Name")

	@cached_property
	def GroupId(self):
		return self.__getattr__("GroupId")

	@cached_property
	def ConfiguredIn(self):
		return self.__getattr__("ConfiguredIn")





#---------------------------------------------------------------
# This file is generated! Dont make any changes in this file.
#---------------------------------------------------------------
from .base import *
from .cached_property import cached_property



class SessionInfoHandler(BaseGetHandler):
	'''Session Infos

	Methods
	-------
		createEntry()
			Creates a new SessionInfo entry object.
	
	Inherited Methods
	-----------------
		get (BaseGetHandler)
	'''
	

	def createEntry(self):
		'''Creates a new SessionInfo entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return SessionInfo(self)


class SessionInfo(BaseEntry):
	'''Model with login session infos

	Attributes [writable]
	---------------------
		RootGroupId : integer
			Root group REST ID
		RootGroupName : string
			Root group name
		LoginUserName : string
			Login user name
		DisplayName : string
			Administratos display name
		Rights : array from {ParamType}
			Rights
		ExpiresIn : integer
			Expires In
		SemVersion : string
			SEM Version
		SemBuildNr : string
			SEM Build Nr.
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"RootGroupId" : 0,
			"RootGroupName" : "",
			"LoginUserName" : "",
			"DisplayName" : "",
			"Rights" : [],
			"ExpiresIn" : 0,
			"SemVersion" : "",
			"SemBuildNr" : "",
		}

	@cached_property
	def RootGroupId(self):
		return self.__getattr__("RootGroupId")

	@cached_property
	def RootGroupName(self):
		return self.__getattr__("RootGroupName")

	@cached_property
	def LoginUserName(self):
		return self.__getattr__("LoginUserName")

	@cached_property
	def DisplayName(self):
		return self.__getattr__("DisplayName")

	@cached_property
	def Rights(self):
		return self.__getattr__("Rights")

	@cached_property
	def ExpiresIn(self):
		return self.__getattr__("ExpiresIn")

	@cached_property
	def SemVersion(self):
		return self.__getattr__("SemVersion")

	@cached_property
	def SemBuildNr(self):
		return self.__getattr__("SemBuildNr")


class SemInfoHandler(BaseGetHandler):
	'''System Information

	Methods
	-------
		createEntry()
			Creates a new SemInfo entry object.
	
	Inherited Methods
	-----------------
		get (BaseGetHandler)
	'''
	

	def createEntry(self):
		'''Creates a new SemInfo entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return SemInfo(self)


class SemInfo(BaseEntry):
	'''Info from Management Server 

	Attributes [read-only]
	----------------------
		StartTime : time
			Start Time from SEM
		LicensedManagedUnits : integer
			Licensed Managed Units
			Enum Values: 4294967295=unlimited
		UsedManagedUnits : integer
			Used Managed Units
		IsLicenseValid : boolean
			Is License Valid
		SerialNumber : string
			License Serial Number
		LicensedVersion : string
			Licensed Version
		SoftwareVersion : string
			Software Version
		BuildNr : string
			Build Nr
		CommitID : string
			Commit ID
		DBDriverVersion : string
			DB Driver Version
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"StartTime" : "",
			"LicensedManagedUnits" : 0,
			"UsedManagedUnits" : 0,
			"IsLicenseValid" : False,
			"SerialNumber" : "",
			"LicensedVersion" : "",
			"SoftwareVersion" : "",
			"BuildNr" : "",
			"CommitID" : "",
			"DBDriverVersion" : "",
		}

	@cached_property
	def StartTime(self):
		return self.__getattr__("StartTime")

	@cached_property
	def LicensedManagedUnits(self):
		return self.__getattr__("LicensedManagedUnits")

	@cached_property
	def UsedManagedUnits(self):
		return self.__getattr__("UsedManagedUnits")

	@cached_property
	def IsLicenseValid(self):
		return self.__getattr__("IsLicenseValid")

	@cached_property
	def SerialNumber(self):
		return self.__getattr__("SerialNumber")

	@cached_property
	def LicensedVersion(self):
		return self.__getattr__("LicensedVersion")

	@cached_property
	def SoftwareVersion(self):
		return self.__getattr__("SoftwareVersion")

	@cached_property
	def BuildNr(self):
		return self.__getattr__("BuildNr")

	@cached_property
	def CommitID(self):
		return self.__getattr__("CommitID")

	@cached_property
	def DBDriverVersion(self):
		return self.__getattr__("DBDriverVersion")


class SemSettingsHandler(BaseListGetHandler, BaseListUpdateHandler):
	'''SEM System Settings

	Methods
	-------
		createEntry()
			Creates a new SemSettingEntry entry object.
	
	Inherited Methods
	-----------------
		get (BaseListGetHandler)
		list (BaseListHandler)
		update (BaseListUpdateHandler)
	'''
	

	def createEntry(self):
		'''Creates a new SemSettingEntry entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return SemSettingEntry(self)


class SemSettingEntry(LazyModifiableListEntry):
	'''General settings of Management Server 

	Attributes [read-only]
	----------------------
		Id : integer
			Id
		Name : string
			Name
		Type : integer
			Type
			Enum Values: 1=int, 2=str, 3=bool, 4=encstr
		GroupName : string
			Group Name
		ReadOnly : integer
			Read only

	Attributes [writable]
	---------------------
		Value : string
			Value
	'''

	def __init__(self, getHandler):
		LazyModifiableListEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Name" : "",
			"Type" : 0,
			"GroupName" : "",
			"Value" : "",
			"ReadOnly" : 0,
		}

	@cached_property
	def Id(self):
		return self.__getattr__("Id")

	@cached_property
	def Name(self):
		return self.__getattr__("Name")

	@cached_property
	def Type(self):
		return self.__getattr__("Type")

	@cached_property
	def GroupName(self):
		return self.__getattr__("GroupName")

	@cached_property
	def Value(self):
		return self.__getattr__("Value")

	@cached_property
	def ReadOnly(self):
		return self.__getattr__("ReadOnly")


class SemLicenseHandler(BaseUpdateHandler, BaseGetHandler):
	'''SEM License

	Methods
	-------
		createEntry()
			Creates a new SemLicense entry object.
	
	Inherited Methods
	-----------------
		get (BaseGetHandler)
		update (BaseUpdateHandler)
	'''
	

	def createEntry(self):
		'''Creates a new SemLicense entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return SemLicense(self)


class SemLicense(LazyEntry):
	'''License Setting of Management Server 

	Attributes [read-only]
	----------------------
		IsLicenseValid : boolean
			Is License Valid
		LicensedVersion : string
			Licensed Version
		LicensedManagedUnits : integer
			Licensed Managed Units
			Enum Values: 4294967295=unlimited
		LicensedServerType : integer
			Licensed Server Type
			Enum Values: 0=Primary Server, 2=Backup Server
		MultiFactorAuthLicense : boolean
			Multi-factor Authentication License
		TimeLeftTestVersion : integer
			Time Left Test Version
		SoftwareVersion : string
			Software Version
		BuildNr : string
			Build Nr

	Attributes [writable]
	---------------------
		SerialNumber : string
			Serial Number
		ActivationKey : string
			Activation Key
	'''

	def __init__(self, getHandler):
		LazyEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"SerialNumber" : "",
			"ActivationKey" : "",
			"IsLicenseValid" : False,
			"LicensedVersion" : "",
			"LicensedManagedUnits" : 0,
			"LicensedServerType" : 0,
			"MultiFactorAuthLicense" : False,
			"TimeLeftTestVersion" : 0,
			"SoftwareVersion" : "",
			"BuildNr" : "",
		}

	@cached_property
	def SerialNumber(self):
		return self.__getattr__("SerialNumber")

	@cached_property
	def ActivationKey(self):
		return self.__getattr__("ActivationKey")

	@cached_property
	def IsLicenseValid(self):
		return self.__getattr__("IsLicenseValid")

	@cached_property
	def LicensedVersion(self):
		return self.__getattr__("LicensedVersion")

	@cached_property
	def LicensedManagedUnits(self):
		return self.__getattr__("LicensedManagedUnits")

	@cached_property
	def LicensedServerType(self):
		return self.__getattr__("LicensedServerType")

	@cached_property
	def MultiFactorAuthLicense(self):
		return self.__getattr__("MultiFactorAuthLicense")

	@cached_property
	def TimeLeftTestVersion(self):
		return self.__getattr__("TimeLeftTestVersion")

	@cached_property
	def SoftwareVersion(self):
		return self.__getattr__("SoftwareVersion")

	@cached_property
	def BuildNr(self):
		return self.__getattr__("BuildNr")


class AdminsHandler(BaseListFindHandler, BaseListUpdateHandler, BaseListInsertHandler, BaseListDeleteHandler):
	'''Management of SEM administrators

	Methods
	-------
		createEntry()
			Creates a new Admin entry object.
	
	Inherited Methods
	-----------------
		delete (BaseListDeleteHandler)
		find (BaseListFindHandler)
		get (BaseListGetHandler)
		insert (BasListInsertHandler)
		list (BaseListHandler)
		update (BaseListUpdateHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "sem-mgm/{groupid}/admins".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid

	def createEntry(self):
		'''Creates a new Admin entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return Admin(self, self._groupid)


class Admin(LazyModifiableListEntry):
	'''Parameters of SEM administror

	Attributes [read-only]
	----------------------
		LastSerialNumber : string
			Last incoming certificate serial number 
		LastLogin : time
			Last login
		LastLogout : time
			Last logout
		LoginErrors : integer
			Number of login errors
		ModifiedOn : time
			Modified on
		ModifiedBy : string
			Modified by

	Attributes [writable]
	---------------------
		Name : string
			Name
		Description : string
			Description
		Password : string
			Password
		State : integer
			State
			Enum Values: 0=denied, 1=granted, 2=locked
		AdminGroup : string or integer from {Controller}
			Administrator Group
		ForceTLS : boolean
			Force TLS connection
		ForceCert : boolean
			Force TLS with certificate
		CertificateSerialNumber : string
			Certificate serial number 
		ManagementIPAddrs : string
			Management IP addresses
	'''

	def __init__(self, getHandler, groupid):
		LazyModifiableListEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Name" : "",
			"Description" : "",
			"Password" : "",
			"State" : "granted",
			"AdminGroup" : "",
			"ForceTLS" : False,
			"ForceCert" : False,
			"CertificateSerialNumber" : "",
			"ManagementIPAddrs" : "",
			"LastSerialNumber" : "",
			"LastLogin" : "",
			"LastLogout" : "",
			"LoginErrors" : 0,
			"ModifiedOn" : "",
			"ModifiedBy" : "",
		}

	@cached_property
	def Name(self):
		return self.__getattr__("Name")

	@cached_property
	def Description(self):
		return self.__getattr__("Description")

	@cached_property
	def Password(self):
		return self.__getattr__("Password")

	@cached_property
	def State(self):
		return self.__getattr__("State")

	@cached_property
	def AdminGroup(self):
		return self.__getattr__("AdminGroup")

	@cached_property
	def ForceTLS(self):
		return self.__getattr__("ForceTLS")

	@cached_property
	def ForceCert(self):
		return self.__getattr__("ForceCert")

	@cached_property
	def CertificateSerialNumber(self):
		return self.__getattr__("CertificateSerialNumber")

	@cached_property
	def ManagementIPAddrs(self):
		return self.__getattr__("ManagementIPAddrs")

	@cached_property
	def LastSerialNumber(self):
		return self.__getattr__("LastSerialNumber")

	@cached_property
	def LastLogin(self):
		return self.__getattr__("LastLogin")

	@cached_property
	def LastLogout(self):
		return self.__getattr__("LastLogout")

	@cached_property
	def LoginErrors(self):
		return self.__getattr__("LoginErrors")

	@cached_property
	def ModifiedOn(self):
		return self.__getattr__("ModifiedOn")

	@cached_property
	def ModifiedBy(self):
		return self.__getattr__("ModifiedBy")


class AdminGroupsHandler(BaseListFindHandler):
	'''Management of administrator groups

	Methods
	-------
		createEntry()
			Creates a new AdminGroup entry object.
	
	Inherited Methods
	-----------------
		find (BaseListFindHandler)
		get (BaseListGetHandler)
		list (BaseListHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "sem-mgm/{groupid}/admin-groups".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid

	def createEntry(self):
		'''Creates a new AdminGroup entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return AdminGroup(self, self._groupid)


class AdminGroup(LazyListEntry):
	'''Parameters of SEM administror group

	Attributes [read-only]
	----------------------
		GroupId : integer
			Group ID
		ConfiguredIn : string
			Configured in
		ModifiedOn : time
			Modified on
		ModifiedBy : string
			Modified by

	Attributes [writable]
	---------------------
		Name : string
			Name
		Description : string
			Description
	'''

	def __init__(self, getHandler, groupid):
		LazyListEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Name" : "",
			"Description" : "",
			"GroupId" : 0,
			"ConfiguredIn" : "",
			"ModifiedOn" : "",
			"ModifiedBy" : "",
		}

	@cached_property
	def Name(self):
		return self.__getattr__("Name")

	@cached_property
	def Description(self):
		return self.__getattr__("Description")

	@cached_property
	def GroupId(self):
		return self.__getattr__("GroupId")

	@cached_property
	def ConfiguredIn(self):
		return self.__getattr__("ConfiguredIn")

	@cached_property
	def ModifiedOn(self):
		return self.__getattr__("ModifiedOn")

	@cached_property
	def ModifiedBy(self):
		return self.__getattr__("ModifiedBy")


class SoftwareUpdateListsHandler(BaseListFindHandler):
	'''Management of software update lists

	Methods
	-------
		createEntry()
			Creates a new GetIdNameGroupList entry object.
	
	Inherited Methods
	-----------------
		find (BaseListFindHandler)
		get (BaseListGetHandler)
		list (BaseListHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "sem-mgm/sw-update-lists/{groupid}".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid

	def createEntry(self):
		'''Creates a new GetIdNameGroupList entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return GetIdNameGroupList(self, self._groupid)


class GetIdNameGroupList(ListEntry):
	'''Model GetIdNameGroupList

	Attributes [read-only]
	----------------------
		Id : integer
			REST ID of the entry
		GroupId : integer
			Group ID of the entry
		ConfiguredIn : string
			Group name of the entry

	Attributes [writable]
	---------------------
		Name : string
			Name of the entry
	'''

	def __init__(self, getHandler, groupid):
		ListEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Name" : "",
			"GroupId" : 0,
			"ConfiguredIn" : "",
		}

	@cached_property
	def Id(self):
		return self.__getattr__("Id")

	@cached_property
	def Name(self):
		return self.__getattr__("Name")

	@cached_property
	def GroupId(self):
		return self.__getattr__("GroupId")

	@cached_property
	def ConfiguredIn(self):
		return self.__getattr__("ConfiguredIn")


class StateDBReplicationHandler(BaseGetHandler):
	'''State Replication Management

	Methods
	-------
		createEntry()
			Creates a new StateDBReplication entry object.
	
	Inherited Methods
	-----------------
		get (BaseGetHandler)
	'''
	

	def createEntry(self):
		'''Creates a new StateDBReplication entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return StateDBReplication(self)


class StateDBReplication(BaseEntry):
	'''State Database Replication

	Attributes [read-only]
	----------------------
		State : integer
			State
			Enum Values: 0=off, 1=disconnected, 2=init, 3=up-to-date, 4=init-update, 5=work, 6=delete, 7=db-error
		ReplicatedRecords : integer
			Replicated Records
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"State" : 0,
			"ReplicatedRecords" : 0,
		}

	@cached_property
	def State(self):
		return self.__getattr__("State")

	@cached_property
	def ReplicatedRecords(self):
		return self.__getattr__("ReplicatedRecords")


class BackupServersHandler(BaseListHandler):
	'''Backup Servers

	Methods
	-------
		createEntry()
			Creates a new BackupServerListEntry entry object.
	
	Inherited Methods
	-----------------
		list (BaseListHandler)
	'''
	

	def createEntry(self):
		'''Creates a new BackupServerListEntry entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return BackupServerListEntry(self)


class BackupServerListEntry(ListEntry):
	'''Parameters of a backup server

	Attributes [read-only]
	----------------------
		OnlineState : integer
			Online State
			Enum Values: 0=Offline, 1=Online
		LastReceiveTime : time
			Last receive time
		ReplState : integer
			Replication State
			Enum Values: 0=Offline, 1=Replication Accounting, 11=Replication Groups, 12=Replication Logs, 13=Replication VPN/GWs, 14=Replication RADIUS Configuration, 15=Replication RADIUS Users, 16=Replication Accounting, 17=Replication Error Counters, 18=Replication RADIUS Group Settings, 19=Replication Certificates, 20=Replication CA Certificates, 21=Replication CRLs, 22=Replication Policy Rules, 23=Replication Policy Parameters, 24=Replication RADIUS Dictionary Attributes, 25=Replication RADIUS Dictionaries, 26=Replication CMP Data, 27=Replication User Templates, 28=Replication Drafts, 29=Replication Packages, 30=Replication Package Access, 31=Replication Software Download Infos, 32=Replication Authentication Codes, 33=Replication Users, 34=Replication Scripts, 35=Replication Profiles, 36=Replication Messages to Primary, 37=Replication Package Files, 38=Replication PKCS#12 Files, 39=Replication Configuration Files, 40=Replication Client Files, 41=Replication File Configuration, 42=Replication Console Profiles, 47=Replication Licenses, 48=Replication Subscriptions, 49=Replication Client Audit Logs, 50=Up-To-Date, 60=Disconnected, 101=Error: Time different greater than one hour between Backup and Primary Server, 102=Error: Duplicate Serial Number, 103=Error: Authentication failed, 104=Error: Backup Licensed Units < Primary Licensed Units, 105=Error: Primary Version < Backup Version, 106=Error: Backup license without NCP Multi-Factor Authentication option

	Attributes [writable]
	---------------------
		Id : integer
			Id of the entry
		Name : string
			Name
		IPAddress : string
			IP Address
	'''

	def __init__(self, getHandler):
		ListEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Name" : "",
			"IPAddress" : "",
			"OnlineState" : 0,
			"LastReceiveTime" : "",
			"ReplState" : 0,
		}

	@cached_property
	def Id(self):
		return self.__getattr__("Id")

	@cached_property
	def Name(self):
		return self.__getattr__("Name")

	@cached_property
	def IPAddress(self):
		return self.__getattr__("IPAddress")

	@cached_property
	def OnlineState(self):
		return self.__getattr__("OnlineState")

	@cached_property
	def LastReceiveTime(self):
		return self.__getattr__("LastReceiveTime")

	@cached_property
	def ReplState(self):
		return self.__getattr__("ReplState")


class SemScriptsHandler(BaseListFindHandler, BaseListUpdateHandler, BaseListInsertHandler, BaseListDeleteHandler):
	'''SEM Scripts

	Methods
	-------
		createEntry()
			Creates a new SemScript entry object.
	
	Inherited Methods
	-----------------
		delete (BaseListDeleteHandler)
		find (BaseListFindHandler)
		get (BaseListGetHandler)
		insert (BasListInsertHandler)
		list (BaseListHandler)
		update (BaseListUpdateHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "sem-mgm/{groupid}/sem-scripts".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid

	def createEntry(self):
		'''Creates a new SemScript entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return SemScript(self, self._groupid)


class SemScript(LazyModifiableListEntry):
	'''Parameters of SEM script

	Attributes [read-only]
	----------------------
		GroupId : integer
			Group ID
		ConfiguredIn : string
			Configured in
		ModifiedOn : time
			Modified on
		ModifiedBy : string
			Modified by

	Attributes [writable]
	---------------------
		Name : string
			Name
		Action : integer
			Execute script manual or by task
			Enum Values: 0=Task, 1=Manual
		Enabled : boolean
			Script enabled
		Script : string
			Script
		MenueItemConsole : boolean
			Create console menue item
		MenueItemWebUI : boolean
			Create Web UI menue item
		Hotkey : string
			Hotkey
		McAction : integer
			Replace funtion of the console
			Enum Values: 0=default, 65539=ClientsCreate, 131075=ClientsDelete, 65850=RadiusUserConfigCreate, 131386=RadiusUserConfigDelete
		InheritedToSubgroups : boolean
			Entry inherited to subgroups
	'''

	def __init__(self, getHandler, groupid):
		LazyModifiableListEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Name" : "",
			"Action" : 0,
			"Enabled" : True,
			"Script" : "",
			"MenueItemConsole" : False,
			"MenueItemWebUI" : False,
			"Hotkey" : "",
			"McAction" : 0,
			"InheritedToSubgroups" : False,
			"GroupId" : 0,
			"ConfiguredIn" : "",
			"ModifiedOn" : "",
			"ModifiedBy" : "",
		}

	@cached_property
	def Name(self):
		return self.__getattr__("Name")

	@cached_property
	def Action(self):
		return self.__getattr__("Action")

	@cached_property
	def Enabled(self):
		return self.__getattr__("Enabled")

	@cached_property
	def Script(self):
		return self.__getattr__("Script")

	@cached_property
	def MenueItemConsole(self):
		return self.__getattr__("MenueItemConsole")

	@cached_property
	def MenueItemWebUI(self):
		return self.__getattr__("MenueItemWebUI")

	@cached_property
	def Hotkey(self):
		return self.__getattr__("Hotkey")

	@cached_property
	def McAction(self):
		return self.__getattr__("McAction")

	@cached_property
	def InheritedToSubgroups(self):
		return self.__getattr__("InheritedToSubgroups")

	@cached_property
	def GroupId(self):
		return self.__getattr__("GroupId")

	@cached_property
	def ConfiguredIn(self):
		return self.__getattr__("ConfiguredIn")

	@cached_property
	def ModifiedOn(self):
		return self.__getattr__("ModifiedOn")

	@cached_property
	def ModifiedBy(self):
		return self.__getattr__("ModifiedBy")


class TasksHandler(BaseListFindHandler, BaseListUpdateHandler, BaseListInsertHandler, BaseListDeleteHandler):
	'''Tasks

	Methods
	-------
		createEntry()
			Creates a new Task entry object.
		start(Script, Arguments=None, GroupId=None, ConfiguredIn=None, ModifiedOn=None, ModifiedBy=None)
			Start a task from a script imediately
	
	Inherited Methods
	-----------------
		delete (BaseListDeleteHandler)
		find (BaseListFindHandler)
		get (BaseListGetHandler)
		insert (BasListInsertHandler)
		list (BaseListHandler)
		update (BaseListUpdateHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "sem-mgm/{groupid}/tasks".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid

	def find(self, searchValue, searchKey="Description", throw=False, options={}):
		return BaseListFindHandler.find(self, searchValue, searchKey=searchKey, throw=throw, options=options)
		
	def start(self, Script, Arguments=None, GroupId=None, ConfiguredIn=None, ModifiedOn=None, ModifiedBy=None):
		'''Start a task from a script imediately
			Script : integer
				Rest-ID of the script
			Arguments : string
				List of arguments
			GroupId : integer
				Group ID
			ConfiguredIn : string
				Configured in
			ModifiedOn : time
				Modified on
			ModifiedBy : string
				Modified by
		'''
		return self._callFunction('/start', Script=Script, Arguments=Arguments, GroupId=GroupId, ConfiguredIn=ConfiguredIn, ModifiedOn=ModifiedOn, ModifiedBy=ModifiedBy)

	def createEntry(self):
		'''Creates a new Task entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return Task(self, self._groupid)


class Task(LazyModifiableListEntry):
	'''Parameters of task

	Attributes [read-only]
	----------------------
		Action : integer
			Action
			Enum Values: 1=create client configuration, 2=distribute license key, 13=create Server configuration, 14=License download, 1001=Script execute
		State : integer
			State
			Enum Values: 0=waiting, 1=running, 2=finished
		NextStarttime : time
			NextStarttime
		GroupId : integer
			Group ID
		ConfiguredIn : string
			Configured in
		ModifiedOn : time
			Modified on
		ModifiedBy : string
			Modified by

	Attributes [writable]
	---------------------
		Script : string or integer from {Controller}
			Script
		Description : string
			Description
		Enabled : boolean
			Enabled
		Interval : integer
			Interval
			Enum Values: 0=time, 3=hour, 4=day, 5=week, 6=mon, 7=year, 8=repeat
		Starttime : time
			Starttime
		TimeOffset : integer
			Time Offset
		RepeatTime : string
			Repeat Time
		Args : string
			Args
	'''

	def __init__(self, getHandler, groupid):
		LazyModifiableListEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Action" : "Script execute",
			"Script" : "",
			"Description" : "",
			"Enabled" : True,
			"Interval" : 0,
			"Starttime" : "",
			"TimeOffset" : 0,
			"RepeatTime" : "01:00",
			"State" : 0,
			"NextStarttime" : "",
			"Args" : "",
			"GroupId" : 0,
			"ConfiguredIn" : "",
			"ModifiedOn" : "",
			"ModifiedBy" : "",
		}

	@cached_property
	def Action(self):
		return self.__getattr__("Action")

	@cached_property
	def Script(self):
		return self.__getattr__("Script")

	@cached_property
	def Description(self):
		return self.__getattr__("Description")

	@cached_property
	def Enabled(self):
		return self.__getattr__("Enabled")

	@cached_property
	def Interval(self):
		return self.__getattr__("Interval")

	@cached_property
	def Starttime(self):
		return self.__getattr__("Starttime")

	@cached_property
	def TimeOffset(self):
		return self.__getattr__("TimeOffset")

	@cached_property
	def RepeatTime(self):
		return self.__getattr__("RepeatTime")

	@cached_property
	def State(self):
		return self.__getattr__("State")

	@cached_property
	def NextStarttime(self):
		return self.__getattr__("NextStarttime")

	@cached_property
	def Args(self):
		return self.__getattr__("Args")

	@cached_property
	def GroupId(self):
		return self.__getattr__("GroupId")

	@cached_property
	def ConfiguredIn(self):
		return self.__getattr__("ConfiguredIn")

	@cached_property
	def ModifiedOn(self):
		return self.__getattr__("ModifiedOn")

	@cached_property
	def ModifiedBy(self):
		return self.__getattr__("ModifiedBy")


class NotificationsHandler(BaseListHandler):
	'''Management of Notifications

	Methods
	-------
		createEntry()
			Creates a new Notification entry object.
		action()
			Perform action on notification
	
	Inherited Methods
	-----------------
		list (BaseListHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "sem-mgm/{groupid}/notifications".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid
		
	def action(self):
		'''Perform action on notification'''
		return self._callFunction('/action/{id}')

	def createEntry(self):
		'''Creates a new Notification entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return Notification(self, self._groupid)


class Notification(ListEntry):
	'''Parameters of the notifications

	Attributes [read-only]
	----------------------
		Id : integer
			Id
		Priority : integer
			Priority
			Enum Values: 3=error, 1=info, 2=warning
		Reason : integer
			Reason
			Enum Values: 0=none, 1=test version, 2=Invalid SEM license, 4=SEM without 2FA license option, 6=SEM License (PPU) must be verified, 7=SEM License (PPU) is not verified, 8=SEM License (PPU) has been deactivated, 5=Backup Server license, 12=SEM certificate expires, 11=SEM certificate has expired, 16=Backup SEM certificate expires, 15=Backup SEM certificate has expired, 61=Secure Server certificate has expired, 62=Secure Server certificate expires, 14=CA certificate expires, 13=CA certificate has expired, 21=Backup server is offline, 22=Backup Server: repl. error, 23=(V)SES offline, 24=(V)SES repl. error, 25=(V)HAS offline, 26=(V)HAS repl. error, 27=(V)SES need restart, 28=(V)SES need netword restart, 31=Software update is disabled, 41=Subscription expired, 42=Subscription not verified, 43=Subscription must be verified, 44=Subscription expires, 70=Low disk space
		Subject : string
			Subject
		StartTime : time
			Start time
		Arg1 : string
			Arg1
		Arg2 : string
			Arg2
	'''

	def __init__(self, getHandler, groupid):
		ListEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Priority" : 0,
			"Reason" : 0,
			"Subject" : "",
			"StartTime" : "",
			"Arg1" : "",
			"Arg2" : "",
		}

	@cached_property
	def Id(self):
		return self.__getattr__("Id")

	@cached_property
	def Priority(self):
		return self.__getattr__("Priority")

	@cached_property
	def Reason(self):
		return self.__getattr__("Reason")

	@cached_property
	def Subject(self):
		return self.__getattr__("Subject")

	@cached_property
	def StartTime(self):
		return self.__getattr__("StartTime")

	@cached_property
	def Arg1(self):
		return self.__getattr__("Arg1")

	@cached_property
	def Arg2(self):
		return self.__getattr__("Arg2")


class SemLogsHandler(BaseListHandler):
	'''SEM Logs

	Methods
	-------
		createEntry()
			Creates a new SemLogEntry entry object.
	
	Inherited Methods
	-----------------
		list (BaseListHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "sem-mgm/{groupid}/logs".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid

	def createEntry(self):
		'''Creates a new SemLogEntry entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return SemLogEntry(self, self._groupid)


class SemLogEntry(ListEntry):
	'''Parameters of a SEM log entry

	Attributes [read-only]
	----------------------
		Id : integer
			Id
		Message : string
			Message
		Time : time
			Log Time
		ErrorNr : integer
			Error number
		LogType : integer
			Log Type
		Server : string
			Server
	'''

	def __init__(self, getHandler, groupid):
		ListEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Message" : "",
			"Time" : "",
			"ErrorNr" : 0,
			"LogType" : 0,
			"Server" : "",
		}

	@cached_property
	def Id(self):
		return self.__getattr__("Id")

	@cached_property
	def Message(self):
		return self.__getattr__("Message")

	@cached_property
	def Time(self):
		return self.__getattr__("Time")

	@cached_property
	def ErrorNr(self):
		return self.__getattr__("ErrorNr")

	@cached_property
	def LogType(self):
		return self.__getattr__("LogType")

	@cached_property
	def Server(self):
		return self.__getattr__("Server")


class SemTraceLogsHandler(BaseListHandler):
	'''SEM Trace Logs

	Methods
	-------
		createEntry()
			Creates a new SemTraceLogEntry entry object.
	
	Inherited Methods
	-----------------
		list (BaseListHandler)
	'''
	

	def createEntry(self):
		'''Creates a new SemTraceLogEntry entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return SemTraceLogEntry(self)


class SemTraceLogEntry(ListEntry):
	'''Parameters of a SEM trace log entry

	Attributes [read-only]
	----------------------
		Id : integer
			Id
		Message : string
			Message
		Time : time
			Log Time
		LogType : integer
			Log Type
	'''

	def __init__(self, getHandler):
		ListEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Message" : "",
			"Time" : "",
			"LogType" : 0,
		}

	@cached_property
	def Id(self):
		return self.__getattr__("Id")

	@cached_property
	def Message(self):
		return self.__getattr__("Message")

	@cached_property
	def Time(self):
		return self.__getattr__("Time")

	@cached_property
	def LogType(self):
		return self.__getattr__("LogType")


class SemTraceLoggersHandler(BaseListUpdateHandler):
	'''SEM Trace Loggers

	Methods
	-------
		createEntry()
			Creates a new SemTraceLogger entry object.
	
	Inherited Methods
	-----------------
		list (BaseListHandler)
		update (BaseListUpdateHandler)
	'''
	

	def createEntry(self):
		'''Creates a new SemTraceLogger entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return SemTraceLogger(self)


class SemTraceLogger(ModifiableListEntry):
	'''Parameters of a SEM trace logger entry

	Attributes [writable]
	---------------------
		Process : string
			Process
		Logger : string
			Logger
		LogLevel : integer
			Log Level
			Enum Values: 1=error, 2=warning, 3=info, 4=debug, 5=trace, 6=default
	'''

	def __init__(self, getHandler):
		ModifiableListEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Process" : "",
			"Logger" : "",
			"LogLevel" : 0,
		}

	@cached_property
	def Process(self):
		return self.__getattr__("Process")

	@cached_property
	def Logger(self):
		return self.__getattr__("Logger")

	@cached_property
	def LogLevel(self):
		return self.__getattr__("LogLevel")



class GetIdNameList(BaseEntry):
	'''Model GetIdNameList

	Attributes [writable]
	---------------------
		Id : integer
			ID of the entry
		Name : string
			Name of the entry
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Name" : "",
		}

	@cached_property
	def Id(self):
		return self.__getattr__("Id")

	@cached_property
	def Name(self):
		return self.__getattr__("Name")


class SemScriptList(BaseEntry):
	'''List of Sem scripts

	Attributes [read-only]
	----------------------
		GroupId : integer
			Group ID
		ConfiguredIn : string
			Configured in

	Attributes [writable]
	---------------------
		Id : integer
			Id
		Name : string
			Name
		Action : integer
			Execute script manual or by task
			Enum Values: 0=Task, 1=Manual
		Enabled : boolean
			Script enabled
		MenueItemConsole : boolean
			Create console menue item
		MenueItemWebUI : boolean
			Create Web UI menue item
		Hotkey : string
			Hotkey
		McAction : integer
			Replace funtion of the console
			Enum Values: 0=default, 65539=ClientsCreate, 131075=ClientsDelete, 65850=RadiusUserConfigCreate, 131386=RadiusUserConfigDelete
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Name" : "",
			"Action" : 0,
			"Enabled" : True,
			"MenueItemConsole" : False,
			"MenueItemWebUI" : False,
			"Hotkey" : "",
			"McAction" : 0,
			"GroupId" : 0,
			"ConfiguredIn" : "",
		}

	@cached_property
	def Id(self):
		return self.__getattr__("Id")

	@cached_property
	def Name(self):
		return self.__getattr__("Name")

	@cached_property
	def Action(self):
		return self.__getattr__("Action")

	@cached_property
	def Enabled(self):
		return self.__getattr__("Enabled")

	@cached_property
	def MenueItemConsole(self):
		return self.__getattr__("MenueItemConsole")

	@cached_property
	def MenueItemWebUI(self):
		return self.__getattr__("MenueItemWebUI")

	@cached_property
	def Hotkey(self):
		return self.__getattr__("Hotkey")

	@cached_property
	def McAction(self):
		return self.__getattr__("McAction")

	@cached_property
	def GroupId(self):
		return self.__getattr__("GroupId")

	@cached_property
	def ConfiguredIn(self):
		return self.__getattr__("ConfiguredIn")


class StartTaskRequest(BaseEntry):
	'''Parameters of task

	Attributes [read-only]
	----------------------
		GroupId : integer
			Group ID
		ConfiguredIn : string
			Configured in
		ModifiedOn : time
			Modified on
		ModifiedBy : string
			Modified by

	Attributes [writable]
	---------------------
		Script : integer
			Rest-ID of the script
		Arguments : string
			List of arguments
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Script" : 0,
			"Arguments" : "",
			"GroupId" : 0,
			"ConfiguredIn" : "",
			"ModifiedOn" : "",
			"ModifiedBy" : "",
		}

	@cached_property
	def Script(self):
		return self.__getattr__("Script")

	@cached_property
	def Arguments(self):
		return self.__getattr__("Arguments")

	@cached_property
	def GroupId(self):
		return self.__getattr__("GroupId")

	@cached_property
	def ConfiguredIn(self):
		return self.__getattr__("ConfiguredIn")

	@cached_property
	def ModifiedOn(self):
		return self.__getattr__("ModifiedOn")

	@cached_property
	def ModifiedBy(self):
		return self.__getattr__("ModifiedBy")


class SemTraceLoggerUpdate(BaseEntry):
	'''Parameters for settings SEM trace loggers

	Attributes [writable]
	---------------------
		Loggers : array from model {Model}
			Loggers
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)
		# Default values
		self._defaultValues = {
			"Loggers" : [],
		}

	@cached_property
	def Loggers(self):
		return self.__getattr__("Loggers")





#---------------------------------------------------------------
# This file is generated! Dont make any changes in this file.
#---------------------------------------------------------------
from .base import *
from .cached_property import cached_property



class RadiusTemplatesHandler(BaseListFindHandler, BaseListUpdateHandler, BaseListInsertHandler, BaseListDeleteHandler):
	'''Management of RADIUS Templates

	Methods
	-------
		createEntry()
			Creates a new RadiusTemplate entry object.
	
	Inherited Methods
	-----------------
		delete (BaseListDeleteHandler)
		find (BaseListFindHandler)
		get (BaseListGetHandler)
		insert (BasListInsertHandler)
		list (BaseListHandler)
		update (BaseListUpdateHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "radius-mgm/{groupid}/templates".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid

	def createEntry(self):
		'''Creates a new RadiusTemplate entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return RadiusTemplate(self, self._groupid)
	

class RadiusTemplate(LazyModifiableListEntry):
	'''Parameters of RADIUS template

	Attributes [read-only]
	----------------------
		ConfiguredIn : string
			Configured in
		ModifiedOn : time
			Modified on
		ModifiedBy : string
			Modified by

	Attributes [writable]
	---------------------
		Name : string
			Name
		GenerateTotpSecrets : boolean
			Generate T-OTP Secrets
		TOTPlabel : string
			T-OTP Label
		ForceNCPAppAuth : boolean
			Force NCP App Authentication
		Attributes : object(Model)
			Attributes
		InheritedToSubgroups : boolean
			Entry inherited to subgroups
	'''

	def __init__(self, getHandler, groupid):
		LazyModifiableListEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Name" : "",
			"GenerateTotpSecrets" : False,
			"TOTPlabel" : "",
			"ForceNCPAppAuth" : False,
			"Attributes" : [],
			"InheritedToSubgroups" : False,
			"ConfiguredIn" : "",
			"ModifiedOn" : "",
			"ModifiedBy" : "",
		}
	
	@cached_property
	def Name(self):
		return self.__getattr__("Name")
	
	@cached_property
	def GenerateTotpSecrets(self):
		return self.__getattr__("GenerateTotpSecrets")
	
	@cached_property
	def TOTPlabel(self):
		return self.__getattr__("TOTPlabel")
	
	@cached_property
	def ForceNCPAppAuth(self):
		return self.__getattr__("ForceNCPAppAuth")
	
	@cached_property
	def Attributes(self):
		return self.__getattr__("Attributes")
	
	@cached_property
	def InheritedToSubgroups(self):
		return self.__getattr__("InheritedToSubgroups")
	
	@cached_property
	def ConfiguredIn(self):
		return self.__getattr__("ConfiguredIn")
	
	@cached_property
	def ModifiedOn(self):
		return self.__getattr__("ModifiedOn")
	
	@cached_property
	def ModifiedBy(self):
		return self.__getattr__("ModifiedBy")
	

class RadiusUserConfigurationsHandler(BaseListFindHandler, BaseListUpdateHandler, BaseListInsertHandler, BaseListDeleteHandler):
	'''Management of RADIUS Users

	Methods
	-------
		createEntry()
			Creates a new RadiusUserConfig entry object.
		generate_all_configs()
			Generates all RADIUS user configurations in this SEM group
	
	Inherited Methods
	-----------------
		delete (BaseListDeleteHandler)
		find (BaseListFindHandler)
		get (BaseListGetHandler)
		insert (BasListInsertHandler)
		list (BaseListHandler)
		update (BaseListUpdateHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "radius-mgm/{groupid}/user-configs".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid
		
	def generate_all_configs(self):
		'''Generates all RADIUS user configurations in this SEM group'''
		return self._callFunction('/generate-all-configs')

	def createEntry(self):
		'''Creates a new RadiusUserConfig entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return RadiusUserConfig(self, self._groupid)
	

class RadiusUserConfig(LazyModifiableListEntry):
	'''Parameters of RADIUS user configuration

	Attributes [read-only]
	----------------------
		TotpSecret : string
			Time based OTP secret
		TotpURI : string
			Time based OTP URI
		LastChangeTime : time
			Last change time
		LastCreationTime : time
			Last creation time
		ModifiedOn : time
			Modified on
		ModifiedBy : string
			Modified by

	Attributes [writable]
	---------------------
		Name : string
			Name
		Template : string or integer from {Controller}
			Template
		UserName : string
			User Name
		Password : string
			Password
		Attributes : object(Model)
			Attributes
		NotBefore : time
			Not before
		NotAfter : time
			Not after

	Methods
	-------
		generate_config()
			Generates a RADIUS user configuration
	'''

	def __init__(self, getHandler, groupid):
		LazyModifiableListEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Name" : "",
			"Template" : "",
			"UserName" : "",
			"Password" : "",
			"TotpSecret" : "",
			"TotpURI" : "",
			"Attributes" : [],
			"NotBefore" : "",
			"NotAfter" : "",
			"LastChangeTime" : "",
			"LastCreationTime" : "",
			"ModifiedOn" : "",
			"ModifiedBy" : "",
		}
	
	@cached_property
	def Name(self):
		return self.__getattr__("Name")
	
	@cached_property
	def Template(self):
		return self.__getattr__("Template")
	
	@cached_property
	def UserName(self):
		return self.__getattr__("UserName")
	
	@cached_property
	def Password(self):
		return self.__getattr__("Password")
	
	@cached_property
	def TotpSecret(self):
		return self.__getattr__("TotpSecret")
	
	@cached_property
	def TotpURI(self):
		return self.__getattr__("TotpURI")
	
	@cached_property
	def Attributes(self):
		return self.__getattr__("Attributes")
	
	@cached_property
	def NotBefore(self):
		return self.__getattr__("NotBefore")
	
	@cached_property
	def NotAfter(self):
		return self.__getattr__("NotAfter")
	
	@cached_property
	def LastChangeTime(self):
		return self.__getattr__("LastChangeTime")
	
	@cached_property
	def LastCreationTime(self):
		return self.__getattr__("LastCreationTime")
	
	@cached_property
	def ModifiedOn(self):
		return self.__getattr__("ModifiedOn")
	
	@cached_property
	def ModifiedBy(self):
		return self.__getattr__("ModifiedBy")
			
	def generate_config(self):
		'''Generates a RADIUS user configuration'''
		return self._callMethod('/generate-config')
	

class RadiusGroupSettingsHandler(BaseUpdateHandler, BaseGetHandler):
	'''Management of RADIUS group settings

	Methods
	-------
		createEntry()
			Creates a new RadiusGroupSettings entry object.
	
	Inherited Methods
	-----------------
		get (BaseGetHandler)
		update (BaseUpdateHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "radius-mgm/{groupid}/group-settings".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid

	def createEntry(self):
		'''Creates a new RadiusGroupSettings entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return RadiusGroupSettings(self, self._groupid)
	

class RadiusGroupSettings(LazyEntry):
	'''Parameters of RADIUS group settings

	Attributes [read-only]
	----------------------
		ConfiguredIn : string
			Configured in
		ModifiedOn : time
			Modified on
		ModifiedBy : string
			Modified by

	Attributes [writable]
	---------------------
		AllowProtPAP : boolean
			Allow protocol PAP
		AllowProtCHAP : boolean
			Allow protocol CHAP
		AllowProtMsChapV1 : boolean
			Allow protocol MSCHAPv1
		AllowProtMsChapV2 : boolean
			Allow protocol MSCHAPv2
		AllowProtEapMd5 : boolean
			Allow protocol EAP-MD5
		AllowProtEapTls : boolean
			Allow protocol EAP-TLS
		AllowProtEapNcpOtp : boolean
			Allow protocol EAP-NCP-OTP
		AllowProtEapMsChapV2 : boolean
			Allow protocol EAP-MSCHAPv2
		MaxErrorCount : integer
			Max. wrong RADIUS logins
		ResetTime : integer
			Reset RADIUIS lock after
		EapTlsCheckMode : integer
			EAP-TLS certificate check
			Enum Values: 0=disabled, 1=userNameEqCN, 2=UserNameEqEMail, 101=CNeqAttributeValue, 102=EMaileqAttributeValue
		TlsCheckAttributeName : string
			Certificate check attribute
		ExAuthNAS : boolean
			Ext. Authenitcation for NAS users
		ExAuthVPN : boolean
			Ext. Authenitcation for VPN users
		ExAuthManual : boolean
			Ext. Authenitcation for manual generated users
		ModifyUsername : integer
			Modify Username
			Enum Values: 0=keepUsername, 1=removeSuffix, 2=removePrefix
		ModifyUsernameString : string
			Modify Username String
		ExtAuthProtocol : integer
			Ext. authentication protocol
			Enum Values: 0=disabled, 1=otp, 2=ldap, 3=radius, 4=kerberos
		ExtAuthHostPrimary : string
			Ext. authentication primary host
		ExtAuthHostBackup : string
			Ext. authentication backup host
		ExtAuthPort : integer
			Ext. authentication port
		ExAuthLdapBindDN : string
			Ext. authentication LDAP bind DN
		ExAuthLdapMemberOf : string
			Ext. authentication LDAP member of
		ExAuthUseTLS : boolean
			Ext. authentication with TLS
		ExAuthDomainName : string
			Ext. authentication domain name
		ExAuthOtpSecret : string
			Ext. authentication OTP secret
		AdvAuthUseNAS : boolean
			NCP Adv. authenitcation for NAS users
		AdvAuthUseVPN : boolean
			NCP Adv. authenitcation for VPN users
		AdvAuthUseManual : boolean
			NCP Adv. authenitcation for manual generated users
		AdvAuthProvider : string
			NCP Adv. authenitcation provider
		AdvAuthUrlPrimary : string
			NCP Adv. authenitcation primary URL
		AdvAuthUrlBackup : string
			NCP Adv. authenitcation backup URL
		AdvAuthProxyHost : string
			NCP Adv. authenitcation proxy host
		AdvAuthProxyPort : integer
			NCP Adv. authenitcation proxy port
		AdvAuthProxyUsername : string
			NCP Adv. authenitcation username
		AdvAuthProxyPassword : string
			NCP Adv. authenitcation password
		AdvAuthPasscodeKind : integer
			NCP Adv. passcode kind
			Enum Values: 0=onlyNum, 1=alpha, 2=numAndAlpha, 3=all
		AdvAuthPasscodeLen : integer
			NCP Adv. passcode length
		TotpUseNAS : boolean
			Time-Based OTP for NAS users
		TotpUseVPN : boolean
			Time-Based OTP for VPN users
		TotpUseManual : boolean
			Time-Based OTP for manual generated users
		InheritedToSubgroups : boolean
			Entry inherited to subgroups
	'''

	def __init__(self, getHandler, groupid):
		LazyEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"AllowProtPAP" : True,
			"AllowProtCHAP" : True,
			"AllowProtMsChapV1" : True,
			"AllowProtMsChapV2" : True,
			"AllowProtEapMd5" : True,
			"AllowProtEapTls" : True,
			"AllowProtEapNcpOtp" : True,
			"AllowProtEapMsChapV2" : True,
			"MaxErrorCount" : 5,
			"ResetTime" : 0,
			"EapTlsCheckMode" : 0,
			"TlsCheckAttributeName" : "",
			"ExAuthNAS" : False,
			"ExAuthVPN" : False,
			"ExAuthManual" : False,
			"ModifyUsername" : 0,
			"ModifyUsernameString" : "",
			"ExtAuthProtocol" : 0,
			"ExtAuthHostPrimary" : "",
			"ExtAuthHostBackup" : "",
			"ExtAuthPort" : 0,
			"ExAuthLdapBindDN" : "",
			"ExAuthLdapMemberOf" : "",
			"ExAuthUseTLS" : False,
			"ExAuthDomainName" : "",
			"ExAuthOtpSecret" : "",
			"AdvAuthUseNAS" : False,
			"AdvAuthUseVPN" : False,
			"AdvAuthUseManual" : False,
			"AdvAuthProvider" : "",
			"AdvAuthUrlPrimary" : "",
			"AdvAuthUrlBackup" : "",
			"AdvAuthProxyHost" : "",
			"AdvAuthProxyPort" : 0,
			"AdvAuthProxyUsername" : "",
			"AdvAuthProxyPassword" : "",
			"AdvAuthPasscodeKind" : 0,
			"AdvAuthPasscodeLen" : 0,
			"TotpUseNAS" : False,
			"TotpUseVPN" : False,
			"TotpUseManual" : False,
			"InheritedToSubgroups" : False,
			"ConfiguredIn" : "",
			"ModifiedOn" : "",
			"ModifiedBy" : "",
		}
	
	@cached_property
	def AllowProtPAP(self):
		return self.__getattr__("AllowProtPAP")
	
	@cached_property
	def AllowProtCHAP(self):
		return self.__getattr__("AllowProtCHAP")
	
	@cached_property
	def AllowProtMsChapV1(self):
		return self.__getattr__("AllowProtMsChapV1")
	
	@cached_property
	def AllowProtMsChapV2(self):
		return self.__getattr__("AllowProtMsChapV2")
	
	@cached_property
	def AllowProtEapMd5(self):
		return self.__getattr__("AllowProtEapMd5")
	
	@cached_property
	def AllowProtEapTls(self):
		return self.__getattr__("AllowProtEapTls")
	
	@cached_property
	def AllowProtEapNcpOtp(self):
		return self.__getattr__("AllowProtEapNcpOtp")
	
	@cached_property
	def AllowProtEapMsChapV2(self):
		return self.__getattr__("AllowProtEapMsChapV2")
	
	@cached_property
	def MaxErrorCount(self):
		return self.__getattr__("MaxErrorCount")
	
	@cached_property
	def ResetTime(self):
		return self.__getattr__("ResetTime")
	
	@cached_property
	def EapTlsCheckMode(self):
		return self.__getattr__("EapTlsCheckMode")
	
	@cached_property
	def TlsCheckAttributeName(self):
		return self.__getattr__("TlsCheckAttributeName")
	
	@cached_property
	def ExAuthNAS(self):
		return self.__getattr__("ExAuthNAS")
	
	@cached_property
	def ExAuthVPN(self):
		return self.__getattr__("ExAuthVPN")
	
	@cached_property
	def ExAuthManual(self):
		return self.__getattr__("ExAuthManual")
	
	@cached_property
	def ModifyUsername(self):
		return self.__getattr__("ModifyUsername")
	
	@cached_property
	def ModifyUsernameString(self):
		return self.__getattr__("ModifyUsernameString")
	
	@cached_property
	def ExtAuthProtocol(self):
		return self.__getattr__("ExtAuthProtocol")
	
	@cached_property
	def ExtAuthHostPrimary(self):
		return self.__getattr__("ExtAuthHostPrimary")
	
	@cached_property
	def ExtAuthHostBackup(self):
		return self.__getattr__("ExtAuthHostBackup")
	
	@cached_property
	def ExtAuthPort(self):
		return self.__getattr__("ExtAuthPort")
	
	@cached_property
	def ExAuthLdapBindDN(self):
		return self.__getattr__("ExAuthLdapBindDN")
	
	@cached_property
	def ExAuthLdapMemberOf(self):
		return self.__getattr__("ExAuthLdapMemberOf")
	
	@cached_property
	def ExAuthUseTLS(self):
		return self.__getattr__("ExAuthUseTLS")
	
	@cached_property
	def ExAuthDomainName(self):
		return self.__getattr__("ExAuthDomainName")
	
	@cached_property
	def ExAuthOtpSecret(self):
		return self.__getattr__("ExAuthOtpSecret")
	
	@cached_property
	def AdvAuthUseNAS(self):
		return self.__getattr__("AdvAuthUseNAS")
	
	@cached_property
	def AdvAuthUseVPN(self):
		return self.__getattr__("AdvAuthUseVPN")
	
	@cached_property
	def AdvAuthUseManual(self):
		return self.__getattr__("AdvAuthUseManual")
	
	@cached_property
	def AdvAuthProvider(self):
		return self.__getattr__("AdvAuthProvider")
	
	@cached_property
	def AdvAuthUrlPrimary(self):
		return self.__getattr__("AdvAuthUrlPrimary")
	
	@cached_property
	def AdvAuthUrlBackup(self):
		return self.__getattr__("AdvAuthUrlBackup")
	
	@cached_property
	def AdvAuthProxyHost(self):
		return self.__getattr__("AdvAuthProxyHost")
	
	@cached_property
	def AdvAuthProxyPort(self):
		return self.__getattr__("AdvAuthProxyPort")
	
	@cached_property
	def AdvAuthProxyUsername(self):
		return self.__getattr__("AdvAuthProxyUsername")
	
	@cached_property
	def AdvAuthProxyPassword(self):
		return self.__getattr__("AdvAuthProxyPassword")
	
	@cached_property
	def AdvAuthPasscodeKind(self):
		return self.__getattr__("AdvAuthPasscodeKind")
	
	@cached_property
	def AdvAuthPasscodeLen(self):
		return self.__getattr__("AdvAuthPasscodeLen")
	
	@cached_property
	def TotpUseNAS(self):
		return self.__getattr__("TotpUseNAS")
	
	@cached_property
	def TotpUseVPN(self):
		return self.__getattr__("TotpUseVPN")
	
	@cached_property
	def TotpUseManual(self):
		return self.__getattr__("TotpUseManual")
	
	@cached_property
	def InheritedToSubgroups(self):
		return self.__getattr__("InheritedToSubgroups")
	
	@cached_property
	def ConfiguredIn(self):
		return self.__getattr__("ConfiguredIn")
	
	@cached_property
	def ModifiedOn(self):
		return self.__getattr__("ModifiedOn")
	
	@cached_property
	def ModifiedBy(self):
		return self.__getattr__("ModifiedBy")
	

class RadiusConfigurationsHandler(BaseListFindHandler):
	'''Management of RADIUS configurations

	Methods
	-------
		createEntry()
			Creates a new RadiusConfiguration entry object.
	
	Inherited Methods
	-----------------
		find (BaseListFindHandler)
		get (BaseListGetHandler)
		list (BaseListHandler)
	'''
	

	def createEntry(self):
		'''Creates a new RadiusConfiguration entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return RadiusConfiguration(self)
	

class RadiusConfiguration(LazyListEntry):
	'''Parameters of RADIUS configurations

	Attributes [writable]
	---------------------
		Name : string
			Name
		Attributes : array from model {Model}
			Attributes
	'''

	def __init__(self, getHandler):
		LazyListEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Name" : "",
			"Attributes" : [],
		}
	
	@cached_property
	def Name(self):
		return self.__getattr__("Name")
	
	@cached_property
	def Attributes(self):
		return self.__getattr__("Attributes")
	

class RadiusDictionariesHandler(BaseListFindHandler):
	'''Management of RADIUS dictionaries

	Methods
	-------
		createEntry()
			Creates a new RadiusDictionary entry object.
	
	Inherited Methods
	-----------------
		find (BaseListFindHandler)
		get (BaseListGetHandler)
		list (BaseListHandler)
	'''
	

	def createEntry(self):
		'''Creates a new RadiusDictionary entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return RadiusDictionary(self)
	

class RadiusDictionary(LazyListEntry):
	'''Parameters of RADIUS dictionaries

	Attributes [writable]
	---------------------
		Name : string
			Name
		Attributes : array from model {Model}
			Attributes
	'''

	def __init__(self, getHandler):
		LazyListEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Name" : "",
			"Attributes" : [],
		}
	
	@cached_property
	def Name(self):
		return self.__getattr__("Name")
	
	@cached_property
	def Attributes(self):
		return self.__getattr__("Attributes")
	

class RadiusClientsHandler(BaseListFindHandler, BaseListUpdateHandler, BaseListInsertHandler, BaseListDeleteHandler):
	'''Management of RADIUS clients

	Methods
	-------
		createEntry()
			Creates a new RadiusClient entry object.
	
	Inherited Methods
	-----------------
		delete (BaseListDeleteHandler)
		find (BaseListFindHandler)
		get (BaseListGetHandler)
		insert (BasListInsertHandler)
		list (BaseListHandler)
		update (BaseListUpdateHandler)
	'''
	

	def createEntry(self):
		'''Creates a new RadiusClient entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return RadiusClient(self)
	

class RadiusClient(LazyModifiableListEntry):
	'''Parameters of RADIUS Client

	Attributes [read-only]
	----------------------
		ModifiedOn : time
			Modified on
		ModifiedBy : string
			Modified by

	Attributes [writable]
	---------------------
		Name : string
			Name
		Enabled : boolean
			Enabled
		IpAddress : string
			IP address
		SharetSecret : string
			Sharet secret
		RadiusDictionary : string or integer from {Controller}
			RADIUS Dictionary
		RadiusConfiguration : string or integer from {Controller}
			RADIUS Configuration
		AllowProtPAP : boolean
			Allow protocol PAP
		AllowProtCHAP : boolean
			Allow protocol CHAP
		AllowProtMsChapV1 : boolean
			Allow protocol MSCHAPv1
		AllowProtMsChapV2 : boolean
			Allow protocol MSCHAPv2
		AllowProtEapMd5 : boolean
			Allow protocol EAP-MD5
		AllowProtEapTls : boolean
			Allow protocol EAP-TLS
		AllowProtEapNcpOtp : boolean
			Allow protocol EAP-NCP-OTP
		AllowProtEapMsChapV2 : boolean
			Allow protocol EAP-MSCHAPv2
	'''

	def __init__(self, getHandler):
		LazyModifiableListEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Name" : "",
			"Enabled" : False,
			"IpAddress" : "",
			"SharetSecret" : "",
			"RadiusDictionary" : "0",
			"RadiusConfiguration" : "0",
			"AllowProtPAP" : True,
			"AllowProtCHAP" : True,
			"AllowProtMsChapV1" : True,
			"AllowProtMsChapV2" : True,
			"AllowProtEapMd5" : True,
			"AllowProtEapTls" : True,
			"AllowProtEapNcpOtp" : True,
			"AllowProtEapMsChapV2" : True,
			"ModifiedOn" : "",
			"ModifiedBy" : "",
		}
	
	@cached_property
	def Name(self):
		return self.__getattr__("Name")
	
	@cached_property
	def Enabled(self):
		return self.__getattr__("Enabled")
	
	@cached_property
	def IpAddress(self):
		return self.__getattr__("IpAddress")
	
	@cached_property
	def SharetSecret(self):
		return self.__getattr__("SharetSecret")
	
	@cached_property
	def RadiusDictionary(self):
		return self.__getattr__("RadiusDictionary")
	
	@cached_property
	def RadiusConfiguration(self):
		return self.__getattr__("RadiusConfiguration")
	
	@cached_property
	def AllowProtPAP(self):
		return self.__getattr__("AllowProtPAP")
	
	@cached_property
	def AllowProtCHAP(self):
		return self.__getattr__("AllowProtCHAP")
	
	@cached_property
	def AllowProtMsChapV1(self):
		return self.__getattr__("AllowProtMsChapV1")
	
	@cached_property
	def AllowProtMsChapV2(self):
		return self.__getattr__("AllowProtMsChapV2")
	
	@cached_property
	def AllowProtEapMd5(self):
		return self.__getattr__("AllowProtEapMd5")
	
	@cached_property
	def AllowProtEapTls(self):
		return self.__getattr__("AllowProtEapTls")
	
	@cached_property
	def AllowProtEapNcpOtp(self):
		return self.__getattr__("AllowProtEapNcpOtp")
	
	@cached_property
	def AllowProtEapMsChapV2(self):
		return self.__getattr__("AllowProtEapMsChapV2")
	
	@cached_property
	def ModifiedOn(self):
		return self.__getattr__("ModifiedOn")
	
	@cached_property
	def ModifiedBy(self):
		return self.__getattr__("ModifiedBy")
	

class RadiusUsersHandler(BaseListFindHandler):
	'''Status of RADIUS users

	Methods
	-------
		createEntry()
			Creates a new RadiusUser entry object.
	
	Inherited Methods
	-----------------
		find (BaseListFindHandler)
		get (BaseListGetHandler)
		list (BaseListHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "radius-mgm/{groupid}/users".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid

	def createEntry(self):
		'''Creates a new RadiusUser entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return RadiusUser(self, self._groupid)
	

class RadiusUser(LazyListEntry):
	'''Parameters of RADIUS user status

	Attributes [read-only]
	----------------------
		Id : integer
			Id
		Username : string
			Username
		GroupName : string
			Group Name
		State : integer
			State
			Enum Values: 0=disabled, 1=enabled, 2=locked
		ErrorCount : integer
			Error Count
		CreationTime : time
			Creation time
		LastLoginTime : time
			Last login time
		LastFailedLoginTime : time
			Last failed login time
		FallBack2FAAuthTime : integer
			Fall Back 2FA end date/time

	Methods
	-------
		disable()
			Disable a RADIUS user
		enable()
			Enable a RADIUS user
		call_2fa_fallback(Duration, Passcode=None)
			Enable 2FA fall back for a RADIUS user
	'''

	def __init__(self, getHandler, groupid):
		LazyListEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Username" : "",
			"GroupName" : "",
			"State" : 0,
			"ErrorCount" : 0,
			"CreationTime" : "",
			"LastLoginTime" : "",
			"LastFailedLoginTime" : "",
			"FallBack2FAAuthTime" : 0,
		}
	
	@cached_property
	def Id(self):
		return self.__getattr__("Id")
	
	@cached_property
	def Username(self):
		return self.__getattr__("Username")
	
	@cached_property
	def GroupName(self):
		return self.__getattr__("GroupName")
	
	@cached_property
	def State(self):
		return self.__getattr__("State")
	
	@cached_property
	def ErrorCount(self):
		return self.__getattr__("ErrorCount")
	
	@cached_property
	def CreationTime(self):
		return self.__getattr__("CreationTime")
	
	@cached_property
	def LastLoginTime(self):
		return self.__getattr__("LastLoginTime")
	
	@cached_property
	def LastFailedLoginTime(self):
		return self.__getattr__("LastFailedLoginTime")
	
	@cached_property
	def FallBack2FAAuthTime(self):
		return self.__getattr__("FallBack2FAAuthTime")
			
	def disable(self):
		'''Disable a RADIUS user'''
		return self._callMethod('/disable')
			
	def enable(self):
		'''Enable a RADIUS user'''
		return self._callMethod('/enable')
			
	def call_2fa_fallback(self, Duration, Passcode=None):
		'''Enable 2FA fall back for a RADIUS user
			Duration : integer
				Duration (sec)
			Passcode : string
				Passcode
		'''
		return self._callMethod('/2fa-fallback', Duration=Duration, Passcode=Passcode)
	

class RadiusSessionsHandler(BaseListDeleteHandler):
	'''Status of RADIUS sessions

	Methods
	-------
		createEntry()
			Creates a new RadiusSession entry object.
	
	Inherited Methods
	-----------------
		delete (BaseListDeleteHandler)
		list (BaseListHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "radius-mgm/{groupid}/sessions".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid

	def createEntry(self):
		'''Creates a new RadiusSession entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return RadiusSession(self, self._groupid)
	

class RadiusSession(ListEntry):
	'''Parameters of RADIUS session

	Attributes [read-only]
	----------------------
		Id : integer
			Id
		Username : string
			Username
		RadiusClient : string
			Radius client
		StartTime : time
			Start time
		Duration : integer
			Duration
		SemServerName : string
			SEM Server Name
	'''

	def __init__(self, getHandler, groupid):
		ListEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Username" : "",
			"RadiusClient" : "",
			"StartTime" : "",
			"Duration" : 0,
			"SemServerName" : "",
		}
	
	@cached_property
	def Id(self):
		return self.__getattr__("Id")
	
	@cached_property
	def Username(self):
		return self.__getattr__("Username")
	
	@cached_property
	def RadiusClient(self):
		return self.__getattr__("RadiusClient")
	
	@cached_property
	def StartTime(self):
		return self.__getattr__("StartTime")
	
	@cached_property
	def Duration(self):
		return self.__getattr__("Duration")
	
	@cached_property
	def SemServerName(self):
		return self.__getattr__("SemServerName")
	

class RadiusAccountingDetailsHandler(BaseListHandler):
	'''Accounting from RADIUS users

	Methods
	-------
		createEntry()
			Creates a new RadiusAccountingDetail entry object.
	
	Inherited Methods
	-----------------
		list (BaseListHandler)
	'''
	
	def __init__(self, api, groupid):
		url = "radius-mgm/{groupid}/accounting-details".format(**vars())
		BaseHandler.__init__(self, api, url)
		self._groupid = groupid

	def createEntry(self):
		'''Creates a new RadiusAccountingDetail entry object.
			Call createEntry to get an empty object for insert() or update().
			createEntry is called by get()'''
		return RadiusAccountingDetail(self, self._groupid)
	

class RadiusAccountingDetail(ListEntry):
	'''Parameters of RADIUS accounting detail

	Attributes [read-only]
	----------------------
		Username : string
			Username
		GroupName : string
			Group Name
		StartTime : time
			State
		ConnectionTime : integer
			Connection Time
		RxBytes : integer
			Rx Bytes
		TxBytes : integer
			Tx Bytes
		RadiusClient : string
			Radius Client
		CalledStationId : string
			Called Station ID
		CallingStationId : string
			Calling Station ID
		NasId : string
			NAS ID
		NasPortType : integer
			NAS Port Type
		ServerEndpoint : string
			Server Endpoint
		ClientEndpoint : string
			Client Endpoint
		FramedIpAddress : string
			Client Framed IP Address
		DnsName : string
			DNS Name
	'''

	def __init__(self, getHandler, groupid):
		ListEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Username" : "",
			"GroupName" : "",
			"StartTime" : "",
			"ConnectionTime" : 0,
			"RxBytes" : 0,
			"TxBytes" : 0,
			"RadiusClient" : "",
			"CalledStationId" : "",
			"CallingStationId" : "",
			"NasId" : "",
			"NasPortType" : 0,
			"ServerEndpoint" : "",
			"ClientEndpoint" : "",
			"FramedIpAddress" : "",
			"DnsName" : "",
		}
	
	@cached_property
	def Username(self):
		return self.__getattr__("Username")
	
	@cached_property
	def GroupName(self):
		return self.__getattr__("GroupName")
	
	@cached_property
	def StartTime(self):
		return self.__getattr__("StartTime")
	
	@cached_property
	def ConnectionTime(self):
		return self.__getattr__("ConnectionTime")
	
	@cached_property
	def RxBytes(self):
		return self.__getattr__("RxBytes")
	
	@cached_property
	def TxBytes(self):
		return self.__getattr__("TxBytes")
	
	@cached_property
	def RadiusClient(self):
		return self.__getattr__("RadiusClient")
	
	@cached_property
	def CalledStationId(self):
		return self.__getattr__("CalledStationId")
	
	@cached_property
	def CallingStationId(self):
		return self.__getattr__("CallingStationId")
	
	@cached_property
	def NasId(self):
		return self.__getattr__("NasId")
	
	@cached_property
	def NasPortType(self):
		return self.__getattr__("NasPortType")
	
	@cached_property
	def ServerEndpoint(self):
		return self.__getattr__("ServerEndpoint")
	
	@cached_property
	def ClientEndpoint(self):
		return self.__getattr__("ClientEndpoint")
	
	@cached_property
	def FramedIpAddress(self):
		return self.__getattr__("FramedIpAddress")
	
	@cached_property
	def DnsName(self):
		return self.__getattr__("DnsName")
	


class GetIdNameList(BaseEntry):
	'''Model GetIdNameList

	Attributes [writable]
	---------------------
		Id : integer
			ID of the entry
		Name : string
			Name of the entry
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Name" : "",
		}
	
	@cached_property
	def Id(self):
		return self.__getattr__("Id")
	
	@cached_property
	def Name(self):
		return self.__getattr__("Name")
	

class RadiusTemplateAttribute(BaseEntry):
	'''Model RadiusTemplateAttribute

	Attributes [read-only]
	----------------------
		AttributeType : integer
			Attribute Type
			Enum Values: 1=integer, 2=string, 3=IPv4 address

	Attributes [writable]
	---------------------
		AttributeName : string
			Attribute Name
		AttributeValue : string
			Attribute Value
		UserSpefific : boolean
			User spefific
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"AttributeName" : "",
			"AttributeValue" : "",
			"AttributeType" : 0,
			"UserSpefific" : False,
		}
	
	@cached_property
	def AttributeName(self):
		return self.__getattr__("AttributeName")
	
	@cached_property
	def AttributeValue(self):
		return self.__getattr__("AttributeValue")
	
	@cached_property
	def AttributeType(self):
		return self.__getattr__("AttributeType")
	
	@cached_property
	def UserSpefific(self):
		return self.__getattr__("UserSpefific")
	

class RadiusTemplateAdd(BaseEntry):
	'''Parameters for adding a RADIUS template

	Attributes [writable]
	---------------------
		Name : string
			Name
		GenerateTotpSecrets : boolean
			Generate T-OTP Secrets
		TOTPlabel : string
			T-OTP Label
		ForceNCPAppAuth : boolean
			Force NCP App Authentication
		InheritedToSubgroups : boolean
			Entry inherited to subgroups
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Name" : "",
			"GenerateTotpSecrets" : False,
			"TOTPlabel" : "",
			"ForceNCPAppAuth" : False,
			"InheritedToSubgroups" : False,
		}
	
	@cached_property
	def Name(self):
		return self.__getattr__("Name")
	
	@cached_property
	def GenerateTotpSecrets(self):
		return self.__getattr__("GenerateTotpSecrets")
	
	@cached_property
	def TOTPlabel(self):
		return self.__getattr__("TOTPlabel")
	
	@cached_property
	def ForceNCPAppAuth(self):
		return self.__getattr__("ForceNCPAppAuth")
	
	@cached_property
	def InheritedToSubgroups(self):
		return self.__getattr__("InheritedToSubgroups")
	

class RadiusUserConfigList(BaseEntry):
	'''List of RADIUS user configuration

	Attributes [read-only]
	----------------------
		GroupName : string
			Group Name
		LastChangeTime : time
			Last change time
		LastCreationTime : time
			Last creation time

	Attributes [writable]
	---------------------
		Id : integer
			Id
		Name : string
			Name
		UserName : string
			User Name
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Name" : "",
			"UserName" : "",
			"GroupName" : "",
			"LastChangeTime" : "",
			"LastCreationTime" : "",
		}
	
	@cached_property
	def Id(self):
		return self.__getattr__("Id")
	
	@cached_property
	def Name(self):
		return self.__getattr__("Name")
	
	@cached_property
	def UserName(self):
		return self.__getattr__("UserName")
	
	@cached_property
	def GroupName(self):
		return self.__getattr__("GroupName")
	
	@cached_property
	def LastChangeTime(self):
		return self.__getattr__("LastChangeTime")
	
	@cached_property
	def LastCreationTime(self):
		return self.__getattr__("LastCreationTime")
	

class RadiusAttribute(BaseEntry):
	'''Model RadiusAttribute

	Attributes [read-only]
	----------------------
		AttributeType : integer
			Attribute Type
			Enum Values: 1=integer, 2=string, 3=IPv4 address
		UserSpefific : boolean
			User spefific

	Attributes [writable]
	---------------------
		AttributeName : string
			Attribute Name
		AttributeValue : string
			Attribute Value
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"AttributeName" : "",
			"AttributeValue" : "",
			"AttributeType" : 0,
			"UserSpefific" : False,
		}
	
	@cached_property
	def AttributeName(self):
		return self.__getattr__("AttributeName")
	
	@cached_property
	def AttributeValue(self):
		return self.__getattr__("AttributeValue")
	
	@cached_property
	def AttributeType(self):
		return self.__getattr__("AttributeType")
	
	@cached_property
	def UserSpefific(self):
		return self.__getattr__("UserSpefific")
	

class RadiusUser2FAFallBack(BaseEntry):
	'''Model RadiusUser2FAFallBack

	Attributes [writable]
	---------------------
		Passcode : string
			Passcode
		Duration : integer
			Duration (sec)
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Passcode" : "",
			"Duration" : 0,
		}
	
	@cached_property
	def Passcode(self):
		return self.__getattr__("Passcode")
	
	@cached_property
	def Duration(self):
		return self.__getattr__("Duration")
	

class RadiusClientList(BaseEntry):
	'''Parameters of list RADIUS clients

	Attributes [writable]
	---------------------
		Id : integer
			Id
		Name : string
			Name
		Enabled : boolean
			Enabled
		IpAddress : string
			IP address
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Id" : 0,
			"Name" : "",
			"Enabled" : False,
			"IpAddress" : "",
		}
	
	@cached_property
	def Id(self):
		return self.__getattr__("Id")
	
	@cached_property
	def Name(self):
		return self.__getattr__("Name")
	
	@cached_property
	def Enabled(self):
		return self.__getattr__("Enabled")
	
	@cached_property
	def IpAddress(self):
		return self.__getattr__("IpAddress")
	

class RadiusConfigurationAttribute(BaseEntry):
	'''Parameters of Radius RADIUS configuration entry

	Attributes [writable]
	---------------------
		AttributeName : string
			Attribute Name
		AttributeValue : string
			Attribute Value
		AttributeNr : string
			Attribute Number
		VendorId : integer
			Vendor Id
		AttributeType : integer
			Attribut  Type
			Enum Values: 1=integer, 2=string, 3=IPv4 address
		UserType : integer
			User Type
			Enum Values: 1=only for VPN, 2=only for NAS, 3=both
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"AttributeName" : "",
			"AttributeValue" : "",
			"AttributeNr" : "",
			"VendorId" : 0,
			"AttributeType" : 0,
			"UserType" : "only for VPN",
		}
	
	@cached_property
	def AttributeName(self):
		return self.__getattr__("AttributeName")
	
	@cached_property
	def AttributeValue(self):
		return self.__getattr__("AttributeValue")
	
	@cached_property
	def AttributeNr(self):
		return self.__getattr__("AttributeNr")
	
	@cached_property
	def VendorId(self):
		return self.__getattr__("VendorId")
	
	@cached_property
	def AttributeType(self):
		return self.__getattr__("AttributeType")
	
	@cached_property
	def UserType(self):
		return self.__getattr__("UserType")
	

class RadiusDictionaryAttribute(BaseEntry):
	'''Parameters of Radius Dictionary Attribute

	Attributes [writable]
	---------------------
		Id : integer
			Id
		AttributeName : string
			Attribute Name
		AttributeNr : integer
			Attribute Number
		VendorId : integer
			Vendor Id
		AttributeType : integer
			Attribut  Type
			Enum Values: 1=integer, 2=string, 3=IPv4 address
	'''

	def __init__(self, getHandler):
		BaseEntry.__init__(self, getHandler)		# Default values
		self._defaultValues = {
			"Id" : 0,
			"AttributeName" : "",
			"AttributeNr" : 0,
			"VendorId" : 0,
			"AttributeType" : 0,
		}
	
	@cached_property
	def Id(self):
		return self.__getattr__("Id")
	
	@cached_property
	def AttributeName(self):
		return self.__getattr__("AttributeName")
	
	@cached_property
	def AttributeNr(self):
		return self.__getattr__("AttributeNr")
	
	@cached_property
	def VendorId(self):
		return self.__getattr__("VendorId")
	
	@cached_property
	def AttributeType(self):
		return self.__getattr__("AttributeType")
	



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# NCP Engineering GmbH

import logging
_logger = logging.getLogger(__name__)

# try imports to avoid exceptions when `NcpSemApi` is imported.
# Exceptions will only be thrown when the specific class is used.
try:
    from asyncio import get_running_loop, run
    asyncio_available = True

    import platform
    if platform.system()=='Windows':
        from asyncio import set_event_loop_policy, WindowsSelectorEventLoopPolicy
        set_event_loop_policy(WindowsSelectorEventLoopPolicy())
except (ModuleNotFoundError, ImportError):
    asyncio_available = False
    get_running_loop, run = None, None

# using oktas python library: https://github.com/okta/okta-sdk-python
try:
    from okta.client import Client
    okta_available = True
except ModuleNotFoundError:
    okta_available = False
    Client = None


class OktaQueryParams:
    """Class to bundle parameters for oktas `list_users` api call.

    Attributes:
        Described here:
        users:
            https://developer.okta.com/docs/reference/api/users/#request-parameters-3
        groups:
            https://developer.okta.com/docs/reference/api/groups/#list-groups
        users_in_group:
            https://developer.okta.com/docs/reference/api/groups/#list-group-members

    """
    def __init__(self,
                 limit: str = None,
                 search: str = None,
                 sortBy: str = None,
                 sortOrder: str = None):
        self.limit: str = limit
        self.search: str = search
        self.sortBy: str = sortBy
        self.sortOrder: str = sortOrder

    def as_dict(self) -> dict:
        """Return query parameters as dictionary.

        Returns:
            dict: Dictionary containing *only the set* query paramters.
        """
        if self.sortBy and not self.search:
            _logger.warning("'sortBy' is supported for search queries only")
        if self.sortOrder and not self.search:
            _logger.warning("'sortOrder' is supported for search queries only")
        query_params = {}
        for param in ('limit', 'search', 'sortBy', 'sortOrder'):
            attr = getattr(self, param)
            if attr:
                query_params[param] = attr
        return query_params


class Okta:
    """Okta REST API Class for querying registered users.

    To obtain a token navigate to your Okta dashboard and click on
    `Security -> API` in the left-hand menu.
    Tokens can then be created from the `Tokens`-Tab by clicking on
    `Create token`.
    The domain can be obtained from the profile information in the top right
    section of the Okta dashboard and needs to contain the 'https://' prefix.
    For example: 'https://dev-123456.okta.com'

    More info here:
    https://developer.okta.com/docs/guides/create-an-api-token/main/

    Attributes:
        token (str): Okta API token
        domain (str): `https` prefixed Okta Domain
            example: 
    """
    def __init__(self, token: str, domain: str) -> None:
        if not okta_available:
            raise ModuleNotFoundError(
                    "Okta library not available! "
                    "Try installing it with 'pip install okta'.")
        if not asyncio_available:
            raise ModuleNotFoundError(
                    "Asyncio library not available! "
                    "Please ensure you have Python version 3.7 "
                    "or later installed or install it with "
                    "'pip install asyncio'.")

        config = {
            'orgUrl': domain,
            'token': token
        }
        self.client = Client(config)

    async def list_users_async(
                self, query_params: OktaQueryParams = None) -> list:
        """List all users asyncronously.

        Arguments:
            query_params (OktaQueryParams), optional: object containing
                the desired query parameters.
                Supported attributes: all

        Returns:
            List containing all found users.
        """
        all_users = []
        query_params = query_params.as_dict() if query_params else {}

        users, resp, err = await self.client.list_users(query_params)
        while True:
            if err:
                _logger.error(err.message)
                return []
            if users:
                all_users += users
            if resp.has_next():
                users, err = await resp.next()
            else:
                break
        return [user.as_dict() for user in users]

    def list_users(self, query_params: OktaQueryParams = None) -> list:
        """List all users syncronously.

        This method simply synchronises the `list_users_async` call.

        Arguments:
            query_params (OktaQueryParams): object holding the desired
                query parameters.
                Supported attributes: all

        Returns:
            List containing all found users.
        """
        return self._synchronize(self.list_users_async,
                                 query_params)

    async def list_groups_async(
                self, query_params: OktaQueryParams = None) -> list:
        """List all groups asyncronously.

        Arguments:
            query_params (OktaQueryParams), optional: object containing
                the desired query parameters.
                Supported attributes: all

        Returns:
            List containing all found groups.
        """
        all_groups = []
        query_params = query_params.as_dict() if query_params else {}

        groups, resp, err = await self.client.list_groups(query_params)
        while True:
            if err:
                _logger.error(err.message)
                return []
            if groups:
                all_groups += groups
            if resp.has_next():
                groups, err = await resp.next()
            else:
                break
        return [group.as_dict() for group in groups]

    def list_groups(self, query_params: OktaQueryParams = None) -> list:
        """List all groups syncronously.

        This method simply synchronises the `list_groups_async` call.

        Arguments:
            query_params (OktaQueryParams): object holding the desired
                query parameters.
                Supported attributes: all

        Returns:
            List containing all found groups.
        """
        return self._synchronize(self.list_groups_async,
                                 query_params)
    
    async def list_users_in_group_async(self, group_id: str,
                                query_params: OktaQueryParams = None) -> list:
        """List all users in group $group_id asyncronously.

        Arguments:
            group_id (str): Group Id as obtained from list_groups_async.
            query_params (OktaQueryParams), optional: object containing
                the desired query parameters.
                Supported attributes: 'limit'

        Returns:
            List containing all found users.
        """
        all_users = []
        query_params = query_params.as_dict() if query_params else {}

        users, resp, err = await self.client.list_group_users(group_id, query_params)
        while True:
            if err:
                _logger.error(err.message)
                return []
            if users:
                all_users += users
            if resp.has_next():
                users, err = await resp.next()
            else:
                break
        return [user.as_dict() for user in users]

    def list_users_in_group(self, group_id: str,
                            query_params: OktaQueryParams = None) -> list:
        """List all users in group $group_id syncronously.

        This method simply synchronises the `list_users_in_group_async` call.

        Arguments:
            group_id (str): Group Id as obtained from list_groups_async.
            query_params (OktaQueryParams), optional: object containing
                the desired query parameters.
                Supported attributes: 'limit'

        Returns:
            List containing all found users.
        """
        return self._synchronize(self.list_users_in_group_async,
                                 group_id,
                                 query_params)
    
    def _synchronize(self, function, *args, **kwargs):
        """Synchronize the getter functions."""
        try:
            loop = get_running_loop()
        except RuntimeError:  # There is no current event loop...
            loop = None

        res = function(*args, **kwargs)
        if loop and loop.is_running():
            return loop.run_until_complete(res)
        return run(res)

if __name__ == "__main__":
    import argparse
    from pprint import pprint
    parser = argparse.ArgumentParser("Okta API examples")
    parser.add_argument("token", help="Okta API Token")
    parser.add_argument("domain", help="Okta Domain")
    args = parser.parse_args()

    client = Okta(args.token, args.domain)

    print("EXAMPLE #1: "
        "Getting the first 10 Active Users alphabetically by profile email:")
    query = OktaQueryParams(
        limit=10,
        search='status eq "ACTIVE"',
        sortBy='profile.email'
    )
    users = client.list_users(query)
    for user in users:
        pprint(user, indent=2)

    print("EXAMPLE #2: Getting all users")
    users = client.list_users()
    print(f"Total amount of users: {len(users)}")

    print("EXAMPLE #3: "
        "Getting id,description and name of all groups "
        "created after 2023-01-01:")
    query = OktaQueryParams(search='created gt "2023-01-01T00:00:00.000Z"')
    groups = client.list_groups(query)
    for group in groups:
        pprint({
            "id": group["id"],
            "description": group.get("profile", {}).get("description"),
            "name": group.get("profile", {}).get("name"),
        }, indent=2)

    print("EXAMPLE #4: Getting all groups and show as <id>: <name>")
    groups = client.list_groups()
    for group in groups:
        print(f"{group['id']}: {group.get('profile', {}).get('name')}")

    print("EXAMPLE #6: Getting all members of the first group")
    members = client.list_users_in_group(groups[0]['id'])
    for member in members:
        pprint(member)

    if asyncio_available:
        print("EXAMPLE #7: Using asyncio to use the async methods:")

        async def list_all_users_groups_asynchronously(client: Okta):
            """This simply show how the async methods can be used.
            
            It is not in itself asynchronous,
            first get the users, then the groups.
            """
            users = await client.list_users_async()
            groups = await client.list_users_async()
            return {'users': users, 'groups': groups}
        
        from asyncio import run
        result = run(list_all_users_groups_asynchronously(client))
        pprint(result)
